"""
Tests for plotting.
"""
import unittest
from os.path import abspath, curdir, join
from pathlib import Path

import pyensembl

from agfusion import database, model

data = pyensembl.EnsemblRelease(84, "mouse")
db = database.AGFusionDB(abspath(join(curdir, "agfusion.mus_musculus.84.db")))
db.build = "mus_musculus_84"

data_human = pyensembl.EnsemblRelease(75, "human")
db_human = database.AGFusionDB(abspath(join(curdir, "agfusion.homo_sapiens.75.db")))
db_human.build = "homo_sapiens_75"


class TestSequencePredictionHuman(unittest.TestCase):
    """Test correctly predict human fusions"""

    def test_1(self):
        """
        test CDS and protein correct for junction that is on exon boundaries and
        produces an out-of-frame protein.
        """

        # test the dna and protein coding sequences are correct by comparing
        # with manually generally sequences

        fusion = model.Fusion(
            gene5prime="TMEM87B",
            gene5primejunction=112843681,
            gene3prime="MERTK",
            gene3primejunction=112722768,
            db=db_human,
            pyensembl_data=data_human,
            protein_databases=["pfam"],
            noncanonical=False,
        )

        fusion.save_images("DLG1-BRAF_mouse")

        assert Path(
            "DLG1-BRAF_mouse/ENST00000283206_ENST00000295408.png"
        ).exists(), "Could not save image."



"""Classes for AGFusion database objects
"""
import logging
import sqlite3
import sys
from os.path import abspath, exists, join

from agfusion import utils

logger = logging.getLogger("AGFusion")

try:
    import MySQLdb
except ImportError:
    logger.debug("Could not import MySQLdb.")


class AGFusionDB:
    """
    Class to handle methods around interacting with the AGFusion SQLite3
    database

    name: name of the database
    reference_name
    """

    def __init__(self, database=None, debug=False):

        self.database = abspath(database)
        self.fastas = {}

        self.logger = logging.getLogger("AGFusion")
        if debug:
            self.logger.setLevel(logging.DEBUG)
        else:
            self.logger.setLevel(logging.INFO)
        ch = logging.StreamHandler()
        if debug:
            ch.setLevel(logging.DEBUG)
        else:
            ch.setLevel(logging.INFO)
        formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
        ch.setFormatter(formatter)
        self.logger.addHandler(ch)

        assert exists(self.database), (
            f"AGFusion database at {database} does not exist! "
            "Either run 'agfusion download' or specify the location of the AGFusion "
            "database with the --dbpath flag."
        )

        self.sqlite3_db = sqlite3.connect(abspath(self.database))
        self.sqlite3_cursor = self.sqlite3_db.cursor()
        self.sqlite3_db.commit()

        self.logger.debug("Connected to the database %s", abspath(self.database))

        self.build = ""


class AGFusionDBBManager:
    """
    Class to handle methods managing the SQLite3 database

    name: name of the database
    reference
    """

    def __init__(self, db_dir, species, release, pfam, server):

        self.species = species
        self.release = release
        self.server = server
        self.build = self.species + "_" + str(self.release)
        self.table = utils.ENSEMBL_MYSQL_TABLES[self.species][self.release]

        self.database = join(
            abspath(db_dir),
            "agfusion." + self.species + "." + str(self.release) + ".db",
        )
        self.fastas = {}

        self.logger = logging.getLogger("AGFusion")
        self.logger.setLevel(logging.INFO)
        ch = logging.StreamHandler()
        ch.setLevel(logging.INFO)
        formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
        ch.setFormatter(formatter)
        self.logger.addHandler(ch)

        if not exists(self.database):
            fout = open(abspath(self.database), "a")
            fout.close()
            self.logger.info("Created the database %s", abspath(self.database))

        self.sqlite3_db = sqlite3.connect(abspath(self.database))
        self.sqlite3_cursor = self.sqlite3_db.cursor()
        self.sqlite3_db.commit()

        self.logger.info("Connected to the database %s", abspath(self.database))

        self.ensembl_db = MySQLdb.connect(server, "anonymous")
        self.ensembl_cursor = self.ensembl_db.cursor()
        self.ensembl_cursor.execute("use " + self.table + ";")

        self.logger.info("Connected to the ensembl MySQL server at %s", server)
        self.logger.info("MySQL - use %s", self.table + ";")

        self._check_for_tables()

        # load the PFAM mappings

        self.pfam_mapping = {}

        for line in open(pfam, "r"):
            line = line.rstrip().split("\t")

            pfam_id = line[0]
            pfam_name = line[1]
            pfam_desc = line[3]

            self.pfam_mapping[pfam_id] = {"name": pfam_name, "desc": pfam_desc}

    def _check_for_tables(self):
        """Check if table exists."""

        self.sqlite3_cursor.execute("drop table if exists " + self.build)

        sqlite3_command = (
            "CREATE TABLE "
            + self.build
            + " ("
            + "gene_id text,"
            + "stable_id text,"
            + "entrez_id text,"
            + "gene_name text,"
            + "canonical_transcript_id text);"
        )

        self.logger.info("MySQL - %s", sqlite3_command)

        self.sqlite3_cursor.execute(sqlite3_command)
        self.sqlite3_db.commit()

        # transcript table

        self.sqlite3_cursor.execute("drop table if exists " + self.build + "_transcript")

        sqlite3_command = (
            "CREATE TABLE "
            + self.build
            + "_transcript ("
            + "transcript_id text,"
            + "gene_id text,"
            + "transcript_stable_id text,"
            + "translation_id text);"
        )

        self.logger.info("MySQL - %s", sqlite3_command)

        self.sqlite3_cursor.execute(sqlite3_command)
        self.sqlite3_db.commit()

        # refseq table

        self.sqlite3_cursor.execute("drop table if exists " + self.build + "_refseq")

        sqlite3_command = (
            "CREATE TABLE "
            + self.build
            + "_refseq ("
            + "transcript_id text,"
            + "transcript_stable_id text,"
            + "refseq_id text);"
        )

        self.logger.info("MySQL - %s", sqlite3_command)

        self.sqlite3_cursor.execute(sqlite3_command)
        self.sqlite3_db.commit()

        # protein annotation tables

        for protein_annotation in utils.PROTEIN_ANNOTATIONS:
            self.sqlite3_cursor.execute(
                "drop table if exists " + self.build + "_" + protein_annotation
            )

            sqlite3_command = (
                f"CREATE TABLE {self.build}_{protein_annotation} ("
                "translation_id text,stable_id text,hit_id text,"
                "seq_start integer,seq_end integer,hit_description text,"
                "hit_name text);"
            )

            self.logger.info("MySQL - %s", sqlite3_command)

            self.sqlite3_cursor.execute(sqlite3_command)
            self.sqlite3_db.commit()

    def fetch_gene_names(self):
        """
        Fetch gene by name.
        """

        genes = {}

        # fetch all gene stable ids

        if self.release < 65:
            mysql_command = (
                "SELECT gene.gene_id, gene_stable_id.stable_id, gene.canonical_transcript_id "
                "FROM gene, gene_stable_id WHERE gene.gene_id = gene_stable_id.gene_id;"
            )
        else:
            mysql_command = (
                "SELECT gene.gene_id, gene.stable_id, gene.canonical_transcript_id FROM gene;"
            )

        self.logger.info("MySQL - %s", mysql_command)

        self.ensembl_cursor.execute(mysql_command)
        for g in self.ensembl_cursor.fetchall():
            genes[g[0]] = {
                "stable_id": g[1],
                "canonical_transcript_id": g[2],
                "entrez_id": "",
                "gene_name": "",
            }

        # fetch entrez IDS

        mysql_command = (
            "SELECT gene.gene_id, xref.dbprimary_acc FROM "
            "gene, object_xref, xref,external_db WHERE "
            "gene.gene_id = object_xref.ensembl_id AND "
            "object_xref.ensembl_object_type = 'Gene' AND "
            "object_xref.xref_id = xref.xref_id AND "
            "xref.external_db_id = external_db.external_db_id AND "
            "external_db.db_name = 'EntrezGene';"
        )
        self.logger.info("MySQL - %s", mysql_command)

        self.ensembl_cursor.execute(mysql_command)

        for g in self.ensembl_cursor.fetchall():
            genes[g[0]]["entrez_id"] = g[1]

        # fetch gene names

        if self.build.find("homo_sapiens") != -1:
            gene_name_db = "HGNC"
        elif self.build.find("mus_musculus") != -1:
            gene_name_db = "MGI"
        else:
            self.logger.error("Do not know where to fetch gene names for %s", self.build)
            sys.exit(1)

        mysql_command = (
            """SELECT gene.gene_id, xref.display_label FROM " \
            "gene, object_xref, xref,external_db WHERE " \
            "gene.gene_id = object_xref.ensembl_id AND " \
            "object_xref.ensembl_object_type = 'Gene' AND " \
            "object_xref.xref_id = xref.xref_id AND " \
            "xref.external_db_id = external_db.external_db_id AND " \
            "external_db.db_name = '"""
            + gene_name_db
            + """';"""
        )

        self.logger.info("MySQL - %s", mysql_command)

        self.ensembl_cursor.execute(mysql_command)

        for g, e in self.ensembl_cursor.fetchall():
            genes[g]["gene_name"] = e

        genes = [
            [
                i,
                genes[i]["stable_id"],
                genes[i]["entrez_id"],
                genes[i]["gene_name"],
                genes[i]["canonical_transcript_id"],
            ]
            for i in list(genes.keys())
        ]

        command = (
            f"SQLite - INSERT INTO {self.build} VALUES "
            "(gene_id,stable_id,entrez_id,gene_name,canonical_transcript_id)"
        )
        self.logger.info(command)

        self.sqlite3_cursor.executemany("INSERT INTO " + self.build + " VALUES (?,?,?,?,?)", genes)

        self.sqlite3_db.commit()

    def fetch_transcript_table(self):
        """
        Fetch all transcripts
        """

        if self.release < 65:
            mysql_command = (
                "SELECT transcript.transcript_id, transcript.gene_id, "
                "transcript_stable_id.stable_id FROM transcript, "
                "transcript_stable_id WHERE "
                "transcript.transcript_id = transcript_stable_id.transcript_id;"
            )
        else:
            mysql_command = (
                "SELECT transcript.transcript_id, "
                "transcript.gene_id, transcript.stable_id FROM transcript;"
            )

        self.logger.info("MySQL - %s", mysql_command)
        self.ensembl_cursor.execute(mysql_command)
        transcripts = {}
        for transcript in self.ensembl_cursor.fetchall():
            transcripts[transcript[0]] = {
                "gene_id": transcript[1],
                "transcript_stable_id": transcript[2],
                "translation_id": "",
            }

        # Fetch all transcripts with tranlstions

        mysql_command = (
            "SELECT transcript.transcript_id, "
            "translation.translation_id FROM "
            "transcript, translation WHERE "
            "transcript.transcript_id = translation.transcript_id;"
        )
        self.logger.info("MySQL - %s", mysql_command)
        self.ensembl_cursor.execute(mysql_command)
        for transcript in self.ensembl_cursor.fetchall():
            transcripts[transcript[0]]["translation_id"] = transcript[1]

        transcripts = [
            [
                i,
                transcripts[i]["gene_id"],
                transcripts[i]["transcript_stable_id"],
                transcripts[i]["translation_id"],
            ]
            for i in list(transcripts.keys())
        ]

        log = (
            f"SQLite - INSERT INTO {self.build}_transcript "
            "VALUES (transcript_id,gene_id,transcript_stable_id,translation_id)"
        )
        self.logger.info(log)

        self.sqlite3_cursor.executemany(
            "INSERT INTO " + self.build + "_transcript VALUES (?,?,?,?)", transcripts
        )

        self.sqlite3_db.commit()

    def fetch_refseq_table(self):
        """
        fetch RefSeq IDS
        """

        if self.release < 65:
            mysql_command = (
                "SELECT transcript.transcript_id, transcript_stable_id.stable_id, "
                "xref.display_label FROM "
                "transcript, transcript_stable_id, object_xref, xref, external_db WHERE "
                "transcript.transcript_id = transcript_stable_id.transcript_id and "
                "transcript.transcript_id = object_xref.ensembl_id AND "
                "object_xref.ensembl_object_type = 'Transcript' AND "
                "object_xref.xref_id = xref.xref_id AND "
                "xref.external_db_id = external_db.external_db_id AND "
                "external_db.db_name = 'RefSeq_mRNA';"
            )
        else:
            mysql_command = (
                "SELECT transcript.transcript_id, transcript.stable_id, xref.display_label "
                "FROM transcript, object_xref, xref, external_db WHERE "
                "transcript.transcript_id = object_xref.ensembl_id AND "
                "object_xref.ensembl_object_type = 'Transcript' AND "
                "object_xref.xref_id = xref.xref_id AND "
                "xref.external_db_id = external_db.external_db_id AND "
                "external_db.db_name = 'RefSeq_mRNA';"
            )
        self.logger.info("MySQL - %s", mysql_command)
        self.ensembl_cursor.execute(mysql_command)

        refseqs = [[i[0], i[1], i[2]] for i in self.ensembl_cursor.fetchall()]

        log = (
            f"SQLite - INSERT INTO {self.build}_refseq VALUES "
            "(transcript_id,transcript_stable_id,refseq_id)"
        )
        self.logger.info(log)

        self.sqlite3_cursor.executemany(
            "INSERT INTO " + self.build + "_refseq VALUES (?,?,?)", refseqs
        )

        self.sqlite3_db.commit()

    def fetch_protein_annotation(self):
        """Fetch protein annotation information."""

        for protein_annotation in utils.PROTEIN_ANNOTATIONS:

            if self.release < 70:
                mysql_command = (
                    "SELECT translation.translation_id, translation_stable_id.stable_id, "
                    "protein_feature.hit_name, protein_feature.seq_start, "
                    "protein_feature.seq_end FROM "
                    "analysis, analysis_description, protein_feature, "
                    "translation, translation_stable_id WHERE "
                    "translation_stable_id.translation_id = translation.translation_id AND "
                    "protein_feature.translation_id = translation.translation_id AND "
                    "protein_feature.analysis_id = analysis.analysis_id AND "
                    "analysis.analysis_id = analysis_description.analysis_id AND "
                    "analysis.logic_name = '" + protein_annotation + "';"
                )
            else:
                mysql_command = (
                    "SELECT translation.translation_id, translation.stable_id, "
                    "protein_feature.hit_name, protein_feature.seq_start, "
                    "protein_feature.seq_end, protein_feature.hit_description FROM "
                    "analysis, analysis_description, protein_feature, translation WHERE "
                    "protein_feature.translation_id = translation.translation_id AND "
                    "protein_feature.analysis_id = analysis.analysis_id AND "
                    "analysis.analysis_id = analysis_description.analysis_id AND "
                    "analysis.logic_name = '" + protein_annotation + "';"
                )

            self.logger.info("MySQL - %s", mysql_command)

            self.ensembl_cursor.execute(mysql_command)

            data = list(self.ensembl_cursor.fetchall())

            if self.release < 70:
                data = [i + (None,) for i in data]

            if protein_annotation == "pfam":
                for i, tmp in enumerate(data):
                    tmp = list(tmp)
                    if tmp[2] in self.pfam_mapping:
                        tmp.append(self.pfam_mapping[tmp[2]]["name"])
                        tmp[5] = self.pfam_mapping[tmp[2]]["desc"]
                    else:
                        tmp.append("")
                    data[i] = tmp
            else:
                for i, tmp in enumerate(data):
                    tmp = list(tmp)
                    tmp.append(None)
                    data[i] = tmp

            msg = (
                f"SQLite - INSERT INTO {self.build}_{protein_annotation}"
                " VALUES "
                "(translation_id,stable_id,hit_id,seq_start,seq_end,hit_description,hit_name)"
            )

            self.logger.info(msg)

            self.sqlite3_cursor.executemany(
                "INSERT INTO " + self.build + "_" + protein_annotation + " VALUES (?,?,?,?,?,?,?)",
                data,
            )

            self.sqlite3_db.commit()



""" cli.py
Command line interface
"""

import argparse
import gzip
import shutil
import sys
from os import mkdir, remove
from os.path import exists, join, split
from pathlib import Path

from future.standard_library import install_aliases

install_aliases()
from urllib.error import HTTPError
from urllib.request import urlopen

import pyensembl

from agfusion import __version__, database, exceptions, model, parsers, utils


def list_available_databases():
    """
    List the available databases that can be downloaded.
    #"""

    print("\n")
    print("Species\t\tRelease\t\tShortbut(s)")
    for species, releases in utils.AVAILABLE_ENSEMBL_SPECIES.items():
        for release in releases:
            shortcut = []
            for genome, data in utils.GENOME_SHORTCUTS.items():
                if species in data and release in data:
                    shortcut.append(genome)
            shortcut = ",".join(shortcut)
            print("{species:<10}\t\t{release:<5}\t\t{shortcut:<20}")
    sys.exit(0)


def downloaddb(args):
    """
    Download the AGFusion database from github
    """

    if args.genome is not None:
        if args.genome not in utils.GENOME_SHORTCUTS:
            print("Invalid genome shortcut! Use -a to see available shortcuts.")
            sys.exit(1)
        else:
            species = utils.GENOME_SHORTCUTS[args.genome][0]
            release = str(utils.GENOME_SHORTCUTS[args.genome][1])
    else:
        if args.species is None or args.release is None:
            print("Specify --species and --release or --genome!")
            sys.exit(1)
        species = args.species
        release = str(args.release)

    file_path = join(args.dir, "agfusion." + species + "." + release + ".db.gz")

    print(f"Downloading the AGFusion database to {file_path}...")

    db_url = utils.AGFUSION_DB_URL + species + "." + release + ".db.gz"

    try:
        response = urlopen(db_url)
    except HTTPError:
        print(f"Was unable to download the file {db_url}!")
        sys.exit(1)

    fout = open(file_path, "wb")
    fout.write(response.read())
    fout.close()

    with gzip.open(file_path, "rb") as f_in, open(file_path.replace(".gz", ""), "wb") as f_out:
        shutil.copyfileobj(f_in, f_out)

    remove(file_path)


def annotate(
    gene5prime,
    junction5prime,
    gene3prime,
    junction3prime,
    agfusion_db,
    pyensembl_data,
    args,
    outdir=None,
    colors=None,
    rename=None,
    scale=None,
    batch_out_dir=None,
):
    """
    Annotate the gene fusion
    """

    fusion = model.Fusion(
        gene5prime=gene5prime,
        gene5primejunction=junction5prime,
        gene3prime=gene3prime,
        gene3primejunction=junction3prime,
        db=agfusion_db,
        pyensembl_data=pyensembl_data,
        protein_databases=args.protein_databases,
        noncanonical=args.noncanonical,
    )

    if batch_out_dir is not None:

        outdir = join(
            batch_out_dir,
            fusion.gene5prime.gene.name
            + "-"
            + str(junction5prime)
            + "_"
            + fusion.gene3prime.gene.name
            + "-"
            + str(junction3prime),
        )

    fusion.save_transcript_cdna(out_dir=outdir, middlestar=args.middlestar)
    fusion.save_transcript_cds(out_dir=outdir, middlestar=args.middlestar)
    fusion.save_proteins(out_dir=outdir, middlestar=args.middlestar)

    fusion.save_images(
        out_dir=outdir,
        file_type=args.type,
        scale=scale,
        colors=colors,
        rename=rename,
        fontsize=args.fontsize,
        height=args.height,
        width=args.width,
        dpi=args.dpi,
        no_domain_labels=args.no_domain_labels,
        plot_wt=args.WT,
        exclude=args.exclude_domain,
    )
    fusion.save_tables(out_dir=outdir)


def batch_mode(args, agfusion_db, pyensembl_data, rename, colors):
    """
    Batch mode for annotation fusions from output from a fusion-finding
    algorithm
    """

    if not exists(args.out):
        mkdir(args.out)
    else:
        agfusion_db.logger.warn(f"Output directory {args.out} already exists! Overwriting...")

    if not Path(args.file).exists():
        FileNotFoundError(f"File not found {args.file}")
        sys.exit(1)

    if args.algorithm in parsers.parsers:
        for fusion in parsers.parsers[args.algorithm](args.file, agfusion_db.logger):

            try:
                annotate(
                    gene5prime=fusion["gene5prime"],
                    junction5prime=fusion["gene5prime_junction"],
                    gene3prime=fusion["gene3prime"],
                    junction3prime=fusion["gene3prime_junction"],
                    agfusion_db=agfusion_db,
                    pyensembl_data=pyensembl_data,
                    args=args,
                    colors=colors,
                    rename=rename,
                    scale=None,
                    batch_out_dir=args.out,
                )
            except exceptions.GeneIDException as error:
                agfusion_db.logger.error(error)
            except exceptions.JunctionException as error:
                agfusion_db.logger.error(error)
            except exceptions.TooManyGenesException as error:
                agfusion_db.logger.error(error)
    else:
        available = ",".join(parsers.parsers.keys())
        agfusion_db.logger.error(
            (
                f"'{args.algorithm}' is not an available option for -a! Choose one of the "
                + f"following: {available}."
            )
        )
        sys.exit(1)


def builddb(args):
    """
    Build a AGFusion database
    """

    agfusion_db = database.AGFusionDBBManager(
        args.dir, args.species, args.release, args.pfam, args.server
    )

    agfusion_db.logger.info("Fetching alternative gene names...")

    agfusion_db.fetch_gene_names()

    agfusion_db.logger.info("Fetching transcript tables...")

    agfusion_db.fetch_transcript_table()

    agfusion_db.fetch_refseq_table()

    agfusion_db.logger.info("Fetching protein annotation data...")

    agfusion_db.fetch_protein_annotation()


def add_common_flags(parser):
    """
    Add commaond line flags that are common to multiple sub parsers
    """

    parser.add_argument(
        "-db",
        "--database",
        type=str,
        required=True,
        help="Path to the AGFusion database (e.g. --db /path/to/agfusion.homo_sapiens.87.db)",
    )
    parser.add_argument("-o", "--out", type=str, required=True, help="Directory to save results")
    parser.add_argument(
        "-nc",
        "--noncanonical",
        action="store_true",
        required=False,
        default=False,
        help="(Optional) Include non-canonical gene transcripts "
        + "in the analysis (default False).",
    )
    parser.add_argument(
        "--protein_databases",
        type=str,
        required=False,
        nargs="+",
        default=["pfam", "tmhmm"],
        help="(Optional) Space-delimited list of one or more protein "
        + "feature databases to include when visualizing proteins. "
        + "Options are: pfam, smart, superfamily, tigrfam, pfscan (Prosite_profiles), "
        + "tmhmm (i.e. transmembrane), seg (low_complexity regions), ncoils "
        + "(coiled coil regions), prints, "
        + "pirsf, and signalp (signal peptide regions) "
        + "(default: --protein_databases pfam and tmhmm).",
    )
    parser.add_argument(
        "--recolor",
        type=str,
        required=False,
        default=None,
        action="append",
        help="(Optional) Re-color a domain. Provide the original name of "
        + "the domain then your color (semi-colon delimited, all in "
        + "quotes). Can specify --recolor multiples for each domain. "
        + '(e.g. --color "Pkinase_Tyr;blue" --color "I-set;#006600").',
    )
    parser.add_argument(
        "--rename",
        type=str,
        required=False,
        default=None,
        action="append",
        help="(Optional) Rename a domain. Provide the original name of "
        + "the domain then your new name (semi-colon delimited, "
        + "all in quotes). Can specify --rename multiples for each "
        + 'domain. (e.g. --rename "Pkinase_Tyr;Kinase").',
    )
    parser.add_argument(
        "--exclude_domain",
        type=str,
        required=False,
        default=[],
        nargs="+",
        help="(Optional) Exclude a certain domain(s) from plotting "
        + "by providing a space-separated list of domain names.",
    )
    parser.add_argument(
        "--type",
        type=str,
        required=False,
        default="png",
        help="(Optional) Image file type (png, jpeg, pdf). Default: png",
    )
    parser.add_argument(
        "-w",
        "--width",
        type=int,
        required=False,
        default=10,
        help="(Optional) Image width in inches (default 10).",
    )
    parser.add_argument(
        "-ht",
        "--height",
        type=int,
        required=False,
        default=3,
        help="(Optional) Image file height in inches (default 3).",
    )
    parser.add_argument(
        "--dpi",
        type=int,
        required=False,
        default=None,
        help="(Optional) Dots per inch.",
    )
    parser.add_argument(
        "--fontsize",
        type=int,
        required=False,
        default=12,
        help="(Optional) Fontsize (default 12).",
    )
    parser.add_argument(
        "--WT",
        action="store_true",
        required=False,
        help="(Optional) Include this to plot wild-type architechtures " + "of the 5' and 3' genes",
    )
    parser.add_argument(
        "-ms",
        "--middlestar",
        action="store_true",
        required=False,
        help="(Optional) Insert a * at the junction position for the "
        + "cdna, cds, and protein sequences (default False).",
    )
    parser.add_argument(
        "-ndl",
        "--no_domain_labels",
        action="store_true",
        required=False,
        help="(Optional) Do not label domains.",
    )
    parser.add_argument(
        "--debug",
        default=False,
        action="store_true",
        help="(Optional) Enable debugging logging.",
    )


def build_cli():
    """Build the CLI.

    Returns:
        argparse.Arguments: Parsed arguments.
    """

    parser = argparse.ArgumentParser(description="Annotate Gene Fusion (AGFusion)")
    subparsers = parser.add_subparsers(help="AGFusion programs.", dest="subparser_name")

    annotate_parser = subparsers.add_parser(
        "annotate", help="Annotate and visualize a single fusion."
    )
    annotate_parser.add_argument(
        "-g5", "--gene5prime", type=str, required=True, help="5' gene partner"
    )
    annotate_parser.add_argument(
        "-g3", "--gene3prime", type=str, required=True, help="3' gene partner"
    )
    annotate_parser.add_argument(
        "-j5",
        "--junction5prime",
        type=int,
        required=True,
        help="Genomic location of predicted fuins for the 5' gene partner. "
        + "The 1-based position that is the last nucleotide included in "
        + "the fusion before the junction.",
    )
    annotate_parser.add_argument(
        "-j3",
        "--junction3prime",
        type=int,
        required=True,
        help="Genomic location of predicted fuins for the 3' gene partner. "
        + "The 1-based position that is the first nucleotide included in "
        + "the fusion after the junction.",
    )
    add_common_flags(annotate_parser)
    annotate_parser.add_argument(
        "--scale",
        type=int,
        required=False,
        default=-1,
        help="(Optional) Set maximum width (in amino acids) of the "
        + "figure to rescale the fusion (default: max length of "
        + "fusion product)",
    )

    # batch file parser

    batch_parser = subparsers.add_parser(
        "batch",
        help="Annotate fusions from an output file from a fusion " + "finding algorithm.",
    )
    batch_parser.add_argument(
        "-f",
        "--file",
        type=str,
        required=True,
        help="Output file from fusion-finding algorithm.",
    )
    batch_parser.add_argument(
        "-a",
        "--algorithm",
        type=str,
        required=True,
        help="The fusion-finding algorithm. Can be one of the following: "
        + ", ".join(parsers.parsers.keys())
        + ".",
    )
    add_common_flags(batch_parser)

    # download database

    database_parser = subparsers.add_parser(
        "download", help="Download database for a reference genome."
    )
    database_parser.add_argument(
        "-d",
        "--dir",
        type=str,
        default="",
        help="(Optional) Directory to the database will be downloaded "
        + "to (defaults to current working directory).",
    )
    database_parser.add_argument(
        "-g",
        "--genome",
        type=str,
        default=None,
        help="Specify the genome shortcut (e.g. hg19). To see all"
        + "available shortcuts run 'agfusion download -a'. Either "
        + "specify this or --species and --release.",
    )
    database_parser.add_argument(
        "-s",
        "--species",
        type=str,
        default=None,
        help="The species (e.g. homo_sapiens).",
    )
    database_parser.add_argument(
        "-r", "--release", type=int, default=None, help="The ensembl release (e.g. 87)."
    )
    database_parser.add_argument(
        "-a",
        "--available",
        action="store_true",
        required=False,
        help="List available species and ensembl releases.",
    )

    # build database parser

    build_database_parser = subparsers.add_parser(
        "build", help="Build database for a reference genome."
    )
    build_database_parser.add_argument(
        "-d",
        "--dir",
        type=str,
        required=True,
        help="Directory to write database file to.",
    )
    build_database_parser.add_argument(
        "-s",
        "--species",
        type=str,
        required=True,
        help="The species (e.g. homo_sapiens).",
    )
    build_database_parser.add_argument(
        "-r",
        "--release",
        type=int,
        required=True,
        help="The ensembl release (e.g. 87).",
    )
    build_database_parser.add_argument(
        "--pfam", type=str, required=True, help="File containing PFAM ID mappings."
    )
    build_database_parser.add_argument(
        "--server",
        type=str,
        required=False,
        default="ensembldb.ensembl.org",
        help="(optional) Ensembl server (default ensembldb.ensembl.org)",
    )

    # agfusion version number

    parser.add_argument("-v", "--version", action="version", version=__version__)
    args = parser.parse_args()

    return args


def parse_names_and_colors(args, agfusion_db):
    """parse the re-coloring and re-naming

    Args:
        args (argparse.Namespace): arguments
        agfusion_db (object): AGFusion datapase pointer

    Returns:
        list[dict]: re-mapping of colors and name
    """

    colors = {}
    rename = {}

    if args.rename is not None:
        for i in args.rename:
            pair = i.split(";")

            assert len(pair) == 2, " did not properly specify --rename"

            if pair[0] in rename:
                agfusion_db.logger.warn(f"WARNING - you rename {pair[0]} twice.")

            rename[pair[0]] = pair[1]

    if args.recolor is not None:
        for i in args.recolor:
            pair = i.split(";")

            assert len(pair) == 2, " did not properly specify --colors"

            if pair[0] in colors:
                agfusion_db.logger.warn(f"You specified colors for {pair[0]} twice.")

            if pair[0] in rename:
                colors[rename[pair[0]]] = pair[1]
            else:
                colors[pair[0]] = pair[1]

    return colors, rename


def main():
    """
    Main function for processing command line options
    """

    args = build_cli()

    if args.subparser_name == "build":
        builddb(args)
        sys.exit(0)
    elif args.subparser_name == "download":
        if args.available:
            list_available_databases()
        else:
            downloaddb(args)
        sys.exit(0)

    # single or batch mode

    if not exists(args.out):
        mkdir(args.out)

    # if user does not specify a sqlite database then use the one provided
    # by the package

    db_file = split(args.database)[1]
    species = db_file.split(".")[1]
    release = db_file.split(".")[2]

    assert species in utils.AVAILABLE_ENSEMBL_SPECIES, "unsupported species!"

    agfusion_db = database.AGFusionDB(args.database, debug=args.debug)
    agfusion_db.build = species + "_" + str(release)

    # get the pyensembl data

    pyensembl_data = pyensembl.EnsemblRelease(release, species)

    try:
        pyensembl_data.db
    except ValueError:
        agfusion_db.logger.error(
            "Missing pyensembl data. Run pyensembl install --release "
            + f"{release} --species {species}"
        )
        sys.exit(1)

    colors, rename = parse_names_and_colors(args, agfusion_db)

    # check image file type is valid

    if args.type not in ["png", "pdf", "jpeg"]:
        agfusion_db.logger.error(f"ERROR - provided an incorrect image file type: {args.type}.")
        sys.exit(1)

    if args.subparser_name == "annotate":
        annotate(
            gene5prime=args.gene5prime,
            junction5prime=args.junction5prime,
            gene3prime=args.gene3prime,
            junction3prime=args.junction3prime,
            agfusion_db=agfusion_db,
            pyensembl_data=pyensembl_data,
            args=args,
            outdir=args.out,
            colors=colors,
            rename=rename,
            scale=args.scale,
        )
    elif args.subparser_name == "batch":
        batch_mode(args, agfusion_db, pyensembl_data, rename, colors)



"""
Holds classes for containing information for Gene and Fusion exon and protein information.
"""
import itertools
import os
import re
import sys

from Bio import Seq, SeqIO, SeqRecord, SeqUtils

from agfusion import exceptions, plot, utils


class _Gene:
    """
    Stores the necessary information to specify the architecture of either
    wild-type genes or fusion gene.
    """

    def __init__(
        self,
        genes=None,
        junction=0,
        pyensembl_data=None,
        genome="",
        gene5prime=False,
        db=None,
        noncanonical=False,
    ):
        """
        genes : str or list
            Provide one gene (str) or list of genes (list). In the case of a
            list of genes, it will seach each gene until it finds a hit.

        junction : int

        pyensembl_data : None

        genome : str

        gene5prime : bool

        db : None

        noncanonical : bool
        """

        if isinstance(genes, str):
            genes = [genes]
        elif not isinstance(genes, list):
            db.logger.error(f"parameter 'gene' to _Gene is not string or list: {genes}")
            sys.exit(1)
        if not isinstance(junction, int):
            db.logger.error(f"parameter 'junction' to _Gene is not int: {junction}")
            sys.exit(1)

        self.gene_found = False
        self.provided_transcript = False  # indicates user if user provided transcript

        self.domains = []
        self.transcripts = {}  # stores transcripts and their DB keys
        self.gene = None
        self.pyensembl_data = pyensembl_data
        self.genome = genome
        self.gene5prime = gene5prime
        self.db_pointer = db
        self.noncanonical = noncanonical

        # find the appropriate Ensembl gene ID

        for gene in genes:

            if re.findall("^NM_", gene) or re.findall("^NR_", gene):
                self._search_by_refseq(gene)
            if gene.isdigit() and not self.gene_found:
                self._search_as_entrez(gene)
            if re.findall("(^ENS.*G)", gene.upper()) and not self.gene_found:
                self._search_as_ensembl_id(gene)
            if re.findall("(^ENS.*T)", gene.upper()) and not self.gene_found:
                self._search_as_ensembl_transcript_id(gene)

            if not self.gene_found:

                # else check if it is a gene symbol

                self._search_by_symbol(gene)

                if not self.gene_found:
                    gene = gene.capitalize()
                    self._search_by_symbol(gene)

                if not self.gene_found:
                    gene = gene.upper()
                    self._search_by_symbol(gene)

            if self.gene_found:
                break

        # if gene has not been identified yet

        if not self.gene_found:
            raise exceptions.GeneIDException(",".join(genes))

        # else continue with processing

        self.junction = junction

        if not self.gene.contains(self.gene.contig, self.junction, self.junction):
            raise exceptions.JunctionException(self.gene.name, self.junction)

        # fetch the entrez gene id and canonical transcript id

        sqlite3_command = (
            "SELECT * FROM "
            + self.db_pointer.build
            + ' WHERE stable_id=="'
            + self.gene.gene_id
            + '"'
        )
        self.db_pointer.logger.debug("SQLite - " + sqlite3_command)
        self.db_pointer.sqlite3_cursor.execute(sqlite3_command)
        tmp = self.db_pointer.sqlite3_cursor.fetchall()

        self.gene_id = tmp[0][0]
        self.entrez_id = tmp[0][2]
        self.canonical_transcript_id = tmp[0][4]

        # get the transcripts that will be processed and annotated

        if not noncanonical and not self.provided_transcript:

            # if only want the canonical and did not specify a certain transcript

            sqlite3_command = (
                "SELECT * FROM "
                + self.db_pointer.build
                + '_transcript WHERE transcript_id=="'
                + self.canonical_transcript_id
                + '"'
            )
            self.db_pointer.logger.debug("SQLite - " + sqlite3_command)
            self.db_pointer.sqlite3_cursor.execute(sqlite3_command)
            tmp = db.sqlite3_cursor.fetchall()
            self.transcripts[tmp[0][2]] = tmp[0][0]
        elif not noncanonical and self.provided_transcript:
            pass
        else:
            if self.provided_transcript and noncanonical:
                self.db_pointer.logger.warning(
                    "You provided a transcript ID as well as specified "
                    "--noncanonical flag. Will process all the gene's transcripts."
                )

            # fetch transcript ids

            sqlite3_command = (
                "SELECT * FROM "
                + self.db_pointer.build
                + '_transcript WHERE gene_id=="'
                + self.gene_id
                + '"'
            )
            self.db_pointer.logger.debug("SQLite - " + sqlite3_command)
            self.db_pointer.sqlite3_cursor.execute(sqlite3_command)
            tmp = self.db_pointer.sqlite3_cursor.fetchall()

            for transcript in tmp:
                self.transcripts[transcript[2]] = transcript[0]

    def _search_as_ensembl_transcript_id(self, gene):
        # if it is ensembl transcript id

        if gene in self.pyensembl_data.transcript_ids():
            transcript = self.pyensembl_data.transcript_by_id(gene.upper())
            self.gene = transcript.gene

            self.db_pointer.logger.debug(
                f"Found Ensembl transcript entry for {gene}: {self.gene.id}"
            )
            self.gene_found = True
            self.provided_transcript = True
        else:
            self.db_pointer.logger.debug(f"Found no Ensembl transcript entry for {gene}")

        sqlite3_command = (
            "SELECT * FROM "
            + self.db_pointer.build
            + '_transcript WHERE transcript_stable_id=="'
            + gene
            + '"'
        )
        self.db_pointer.logger.debug("SQLite - " + sqlite3_command)
        self.db_pointer.sqlite3_cursor.execute(sqlite3_command)
        tmp = self.db_pointer.sqlite3_cursor.fetchall()
        self.transcripts[tmp[0][2]] = tmp[0][0]

    def _search_as_ensembl_id(self, gene):
        # if it is ensembl gene id

        if gene in self.pyensembl_data.gene_ids():
            self.gene = self.pyensembl_data.gene_by_id(gene.upper())
            self.db_pointer.logger.debug(f"Found Enrez gene ID entry for {gene}: {self.gene.id}")
            self.gene_found = True
        else:
            self.db_pointer.logger.debug(f"Cannot find Ensembl gene id {gene} in database!")

    def _search_as_entrez(self, gene):
        # if it is an entrez gene ID

        sqlite3_command = (
            "SELECT * FROM " + self.db_pointer.build + ' WHERE entrez_id=="' + gene + '"'
        )
        self.db_pointer.logger.debug("SQLite - " + sqlite3_command)
        self.db_pointer.sqlite3_cursor.execute(sqlite3_command)
        tmp = self.db_pointer.sqlite3_cursor.fetchall()

        if len(tmp) == 1:
            self.gene = self.pyensembl_data.gene_by_id(tmp[0][1])
            self.db_pointer.logger.debug(f"Found Entrez gene ID entry for {gene}: {self.gene.id}")
            self.gene_found = True
        elif len(tmp) > 1:

            self.db_pointer.logger.error(f"Found too many Entrez gene ID entries for {gene}!")
            for i in tmp:
                self.db_pointer.logger.error(i)
            sys.exit(1)
        else:
            self.db_pointer.logger.debug(f"Found no Entrez gene ID entry for {gene}.")

    def _search_by_refseq(self, gene):

        # if it is RefSeq ID

        sqlite3_command = (
            "SELECT * FROM " + self.db_pointer.build + '_refseq WHERE refseq_id=="' + gene + '"'
        )
        self.db_pointer.logger.debug("SQLite - " + sqlite3_command)
        self.db_pointer.sqlite3_cursor.execute(sqlite3_command)
        tmp = self.db_pointer.sqlite3_cursor.fetchall()

        if len(tmp) == 1:
            self.transcripts[tmp[0][1]] = tmp[0][0]
            self.db_pointer.logger.debug(f"Found RefSeq entry for {gene}")
            self.gene_found = True
            self.provided_transcript = True

            # fetch the gene

            transcript = self.pyensembl_data.transcript_by_id(tmp[0][1])
            self.gene = transcript.gene

        elif len(tmp) > 1:
            self.db_pointer.logger.error(f"Found too many RefSeq entries for {gene}!")
            for i in tmp:
                self.db_pointer.logger.error(i)
            sys.exit(1)
        else:
            self.db_pointer.logger.debug(f"Found no RefSeq entry for {gene}")

    def _search_by_symbol(self, gene):
        if gene in self.pyensembl_data.gene_names():
            temp = self.pyensembl_data.genes_by_name(gene)

            if len(temp) > 1:

                # if too many ensembl gene IDs returned
                # use the one located on chromosomes

                for tmp_gene in temp:
                    if tmp_gene.contig in utils.STANDARD_CHROMOSOMES:
                        self.gene = tmp_gene
                        self.gene_found = True

                if self.gene is None:
                    raise exceptions.TooManyGenesException(gene, [x.id for x in temp], self.genome)
            else:
                self.gene = temp[0]
                self.gene_found = True
                self.db_pointer.logger.debug(f"Found gene symbol entry for {gene}: {self.gene.id}")


class Fusion:
    """
    Generates the information needed for the gene fusion
    """

    def __init__(
        self,
        gene5prime=None,
        gene5primejunction=0,
        gene3prime=None,
        gene3primejunction=0,
        db=None,
        pyensembl_data=None,
        protein_databases=None,
        noncanonical=False,
        transcripts_5prime=None,
        transcripts_3prime=None,
    ):
        """
        gene5prime : str

        gene3prime : str

        gene5primejunction : int

        gene3primejunction : int

        db : None

        pyensembl_data : None

        protein_databases : None

        noncanonical : bool

        transcripts_5prime : str

        transcripts_3prime : str
        """

        self.db_pointer = db
        self.pyensembl_data = pyensembl_data

        # get the reference genom

        self.gene5prime = _Gene(
            genes=gene5prime,
            junction=gene5primejunction,
            pyensembl_data=pyensembl_data,
            gene5prime=True,
            db=db,
            noncanonical=noncanonical,
        )

        self.gene3prime = _Gene(
            genes=gene3prime,
            junction=gene3primejunction,
            pyensembl_data=pyensembl_data,
            gene5prime=False,
            db=db,
            noncanonical=noncanonical,
        )

        self.name = self.gene5prime.gene.name + "_" + self.gene3prime.gene.name
        self.name = self.name.replace("/", "-")

        # construct all the fusion transcript combinations

        self.transcripts = {}

        for combo in list(
            itertools.product(
                list(self.gene5prime.transcripts.keys()),
                list(self.gene3prime.transcripts.keys()),
            )
        ):
            transcript1 = pyensembl_data.transcript_by_id(combo[0])
            transcript2 = pyensembl_data.transcript_by_id(combo[1])

            if transcripts_5prime is not None and transcript1.id not in transcripts_5prime:

                continue

            if transcripts_3prime is not None and transcript2.id not in transcripts_3prime:

                continue

            # skip if the junction is outside the range of either transcript

            name = transcript1.id + "_" + transcript2.id

            if not transcript1.contains(
                transcript1.contig, self.gene5prime.junction, self.gene5prime.junction
            ):
                self.transcripts[name] = FusionTranscript(
                    transcript1=transcript1,
                    transcript2=transcript2,
                    gene5prime=self.gene5prime,
                    gene3prime=self.gene3prime,
                    db=db,
                    protein_databases=protein_databases,
                    predict_effect=False,
                )
                self.transcripts[name].effect = "Outside transcript boundry"
                self.transcripts[name].has_coding_potential = False

            elif not transcript2.contains(
                transcript2.contig, self.gene3prime.junction, self.gene3prime.junction
            ):
                self.transcripts[name] = FusionTranscript(
                    transcript1=transcript1,
                    transcript2=transcript2,
                    gene5prime=self.gene5prime,
                    gene3prime=self.gene3prime,
                    db=db,
                    protein_databases=protein_databases,
                    predict_effect=False,
                )
                self.transcripts[name].effect = "Outside transcript boundry"
                self.transcripts[name].has_coding_potential = False

            else:
                self.transcripts[name] = FusionTranscript(
                    transcript1=transcript1,
                    transcript2=transcript2,
                    gene5prime=self.gene5prime,
                    gene3prime=self.gene3prime,
                    db=db,
                    protein_databases=protein_databases,
                )

    def save_images(
        self,
        out_dir="",
        file_type="png",
        fontsize=12,
        dpi=100,
        colors=None,
        rename=None,
        width=8,
        height=2,
        scale=0,
        no_domain_labels=True,
        plot_wt=False,
        exclude=None,
    ):
        """
        Save images of all fusion isoforms and write to file
        Also plot the wild type proteins if desired
        """

        assert file_type in ["png", "pdf"], "provided wrong file type"

        if not os.path.exists(out_dir):
            os.mkdir(out_dir)

        if plot_wt:
            gene5prime_wt = os.path.join(out_dir, self.gene5prime.gene.gene_name.replace("/", "-"))
            gene3prime_wt = os.path.join(out_dir, self.gene3prime.gene.gene_name.replace("/", "-"))
            if not os.path.exists(gene5prime_wt):
                os.mkdir(gene5prime_wt)

            if not os.path.exists(gene3prime_wt):
                os.mkdir(gene3prime_wt)

        for name, transcript in list(self.transcripts.items()):

            if not transcript.has_coding_potential:
                continue

            filename = os.path.join(out_dir, name + "." + file_type)

            pplot = plot.PlotFusionProtein(
                filename=filename,
                width=width,
                height=height,
                dpi=dpi,
                scale=scale,
                fontsize=fontsize,
                colors=colors,
                rename=rename,
                no_domain_labels=no_domain_labels,
                transcript=transcript,
                exclude=exclude,
            )
            pplot.draw()
            pplot.save()

            filename = os.path.join(out_dir, name + ".exon." + file_type)

            pplot = plot.PlotFusionExons(
                transcript=transcript,
                filename=filename,
                width=width,
                height=height,
                dpi=dpi,
                scale=scale,
                fontsize=fontsize,
            )
            pplot.draw()
            pplot.save()

            if plot_wt:

                # plot exons

                filename = os.path.join(
                    gene5prime_wt, transcript.transcript1.id + ".exon." + file_type
                )

                pplot = plot.PlotWTExons(
                    ensembl_transcript=transcript.transcript1,
                    filename=filename,
                    width=width,
                    height=height,
                    dpi=dpi,
                    scale=scale,
                    fontsize=fontsize,
                )
                pplot.draw()
                pplot.save()

                filename = os.path.join(
                    gene3prime_wt, transcript.transcript2.id + ".exon." + file_type
                )

                pplot = plot.PlotWTExons(
                    ensembl_transcript=transcript.transcript2,
                    filename=filename,
                    width=width,
                    height=height,
                    dpi=dpi,
                    scale=scale,
                    fontsize=fontsize,
                )
                pplot.draw()
                pplot.save()

                # plot proteins

                filename = os.path.join(gene5prime_wt, transcript.transcript1.id + "." + file_type)

                pplot = plot.PlotWTProtein(
                    ensembl_transcript=transcript.transcript1,
                    filename=filename,
                    width=width,
                    height=height,
                    dpi=dpi,
                    scale=scale,
                    fontsize=fontsize,
                    colors=colors,
                    rename=rename,
                    no_domain_labels=no_domain_labels,
                    transcript=transcript,
                    exclude=exclude,
                )
                pplot.draw()
                pplot.save()

                filename = os.path.join(gene3prime_wt, transcript.transcript2.id + "." + file_type)

                pplot = plot.PlotWTProtein(
                    ensembl_transcript=transcript.transcript2,
                    filename=filename,
                    width=width,
                    height=height,
                    dpi=dpi,
                    scale=scale,
                    fontsize=fontsize,
                    colors=colors,
                    rename=rename,
                    no_domain_labels=no_domain_labels,
                    transcript=transcript,
                    exclude=exclude,
                )
                pplot.draw()
                pplot.save()

    def save_transcript_cdna(self, out_dir=".", middlestar=False):
        """
        Save the cDNA sequences for all fusion isoforms to a fasta file
        """

        if not os.path.exists(out_dir):
            os.mkdir(out_dir)

        fout = open(os.path.join(out_dir, self.name + "_cdna.fa"), "w")

        for _, transcript in list(self.transcripts.items()):

            if transcript.cdna is not None:

                if middlestar:
                    temp = str(transcript.cdna.seq)
                    temp = (
                        temp[: transcript.transcript_cdna_junction_5prime]
                        + "*"
                        + temp[transcript.transcript_cdna_junction_5prime :]
                    )
                    transcript.cdna.seq = Seq.Seq(temp)

                SeqIO.write(transcript.cdna, fout, "fasta")
            else:
                cdna = SeqRecord.SeqRecord(
                    Seq.Seq(""),
                    id=transcript.name,
                    name=transcript.name,
                    description="No cDNA, fusion junction outside transcript(s) boundary",
                )
                SeqIO.write(cdna, fout, "fasta")

        fout.close()

    def save_transcript_cds(self, out_dir=".", middlestar=False):
        """
        Save the CDS sequences for all fusion isoforms to a fasta file
        """

        if not os.path.exists(out_dir):
            os.mkdir(out_dir)

        # check if any transcripts have coding potential

        n = 0
        for _, transcript in list(self.transcripts.items()):

            if transcript.cds is not None:
                n += 1

        if n == 0:
            self.db_pointer.logger.debug(
                f"The {self.name} fusion does not produce any "
                "protein coding transcripts. No cds.fa file will be written"
            )
            return

        fout = open(os.path.join(out_dir, self.name + "_cds.fa"), "w")

        for _, transcript in list(self.transcripts.items()):

            if transcript.cds is not None:

                if middlestar:
                    temp = str(transcript.cds.seq)
                    temp = (
                        temp[: transcript.transcript_cds_junction_5prime]
                        + "*"
                        + temp[transcript.transcript_cds_junction_5prime :]
                    )
                    transcript.cds.seq = Seq.Seq(temp)

                SeqIO.write(transcript.cds, fout, "fasta")

        fout.close()

    def save_proteins(self, out_dir=".", middlestar=False):
        """
        Save the protein sequences for all fusion isoforms to a fasta file
        """

        if not os.path.exists(out_dir):
            os.mkdir(out_dir)

        # check if any transcripts have coding potential

        n = 0
        for _, transcript in list(self.transcripts.items()):

            if transcript.cds is not None:
                n += 1

        if n == 0:
            self.db_pointer.logger.debug(
                f"The {self.name} fusion does not produce any "
                "protein coding transcripts. No proteins.fa file will be written"
            )
            return

        fout = open(os.path.join(out_dir, self.name + "_protein.fa"), "w")

        for _, transcript in list(self.transcripts.items()):

            if transcript.cds is not None:

                if middlestar:
                    temp = str(transcript.protein.seq)
                    temp = (
                        temp[: transcript.transcript_protein_junction_5prime]
                        + "*"
                        + temp[transcript.transcript_protein_junction_5prime :]
                    )
                    transcript.protein.seq = Seq.Seq(temp)

                SeqIO.write(transcript.protein, fout, "fasta")

        fout.close()

    def save_tables(self, out_dir="."):
        """
        write table of fusion isoforms

        Args:
            out_dir (str, optional): Directory to save results. Defaults to ".".
        """

        fout = open(os.path.join(out_dir, self.name + ".fusion_transcripts.csv"), "w")
        fout.write(
            ",".join(["{}"] * 11).format(
                "5'_gene",
                "3'_gene",
                "5'_transcript",
                "3'_transcript",
                "5'_strand",
                "3'_strand",
                "5'_transcript_biotype",
                "3'_transcript_biotype",
                "Fusion_effect",
                "Protein_length",
                "Protein_weight_(kD)",
            )
            + "\n"
        )

        for _, transcript in list(self.transcripts.items()):

            if transcript.protein_length is None:
                protein_length = "NA"
            else:
                protein_length = transcript.protein_length

            if transcript.molecular_weight is None:
                molecular_weight = "NA"
            else:
                molecular_weight = transcript.molecular_weight

            fout.write(
                ",".join(["{}"] * 11).format(
                    transcript.gene5prime.gene.gene_name,
                    transcript.gene3prime.gene.gene_name,
                    transcript.transcript1.id,
                    transcript.transcript2.id,
                    transcript.transcript1.strand,
                    transcript.transcript2.strand,
                    transcript.transcript1.biotype,
                    transcript.transcript2.biotype,
                    transcript.effect,
                    protein_length,
                    molecular_weight,
                )
                + "\n"
            )
        fout.close()

        # write table containing protein domains for each fusion isoform

        fout = open(os.path.join(out_dir, self.name + ".protein_domains.csv"), "w")
        fout.write(
            ",".join(["{}"] * 11).format(
                "5'_gene",
                "3'_gene",
                "5'_transcript",
                "3'_transcript",
                "5'_strand",
                "3'_strand",
                "Domain_ID",
                "Domain_name",
                "Domain_description",
                "Protein_start",
                "Protein_end",
            )
            + "\n"
        )

        for _, transcript in list(self.transcripts.items()):
            for domain in transcript.domains["fusion"]:
                fout.write(
                    ",".join(["{}"] * 11).format(
                        transcript.gene5prime.gene.gene_name,
                        transcript.gene3prime.gene.gene_name,
                        transcript.transcript1.id,
                        transcript.transcript2.id,
                        transcript.transcript1.strand,
                        transcript.transcript2.strand,
                        domain[0],
                        domain[1],
                        domain[2],
                        domain[3],
                        domain[4],
                    )
                    + "\n"
                )
        fout.close()

        # write tables of exon structure

        fout = open(os.path.join(out_dir, self.name + ".exons.csv"), "w")
        fout.write(
            ",".join(["{}"] * 11).format(
                "5'_gene",
                "3'_gene",
                "5'_transcript",
                "3'_transcript",
                "5'_strand",
                "3'_strand",
                "exon_gene_source",
                "exon_number",
                "exon_chr",
                "exon_start",
                "exon_end",
            )
            + "\n"
        )

        for _, transcript in list(self.transcripts.items()):
            for exon in transcript.gene5prime_exon_intervals:
                fout.write(
                    ",".join(["{}"] * 11).format(
                        transcript.gene5prime.gene.gene_name,
                        transcript.gene3prime.gene.gene_name,
                        transcript.transcript1.id,
                        transcript.transcript2.id,
                        transcript.transcript1.strand,
                        transcript.transcript2.strand,
                        "'5 gene",
                        exon[2],
                        transcript.transcript1.contig,
                        exon[0],
                        exon[1],
                    )
                    + "\n"
                )

            for exon in transcript.gene3prime_exon_intervals:
                fout.write(
                    ",".join(["{}"] * 11).format(
                        transcript.gene5prime.gene.gene_name,
                        transcript.gene3prime.gene.gene_name,
                        transcript.transcript1.id,
                        transcript.transcript2.id,
                        transcript.transcript1.strand,
                        transcript.transcript2.strand,
                        "'3 gene",
                        exon[2],
                        transcript.transcript2.contig,
                        exon[0],
                        exon[1],
                    )
                    + "\n"
                )

        fout.close()


class FusionTranscript:
    """
    Generates the information needed for the gene fusion transctips
    """

    def __init__(
        self,
        transcript1=None,
        transcript2=None,
        gene5prime=None,
        gene3prime=None,
        db=None,
        protein_databases=None,
        predict_effect=True,
    ):

        self.transcript1 = transcript1
        self.transcript2 = transcript2
        self.gene5prime = gene5prime
        self.gene3prime = gene3prime
        self.protein_databases = protein_databases
        self.db_pointer = db

        self.name = self.transcript1.id + "_" + self.transcript2.id
        self.gene_names = self.gene5prime.gene.name + "_" + self.gene3prime.gene.name
        self.gene_names = self.gene_names.replace("/", "-")

        self.cds = None
        self.transcript_cds_junction_5prime = None
        self.transcript_cds_junction_3prime = None
        self.gene5prime_cds_intervals = []
        self.gene3prime_cds_intervals = []
        self.cds_5prime = None
        self.cds_3prime = None

        self.cdna = None
        self.cdna_5prime = None
        self.cdna_3prime = None
        self.transcript_cdna_junction_5prime = None
        self.transcript_cdna_junction_3prime = None
        self.gene5prime_exon_intervals = []
        self.gene3prime_exon_intervals = []

        self.protein = None
        self.protein_length = None
        self.protein_names = None
        self.molecular_weight = None
        self.domains = {transcript1.id: [], transcript2.id: [], "fusion": []}

        self.transcript_protein_junction_5prime = None
        self.transcript_protein_junction_3prime = None

        self.effect_5prime = "exon"
        self.effect_3prime = "exon"
        self.effect = ""
        self.has_coding_potential = False

        if predict_effect:
            self.predict_effect()

    def _annotate(self):
        """
        Annotate the gene fusion's protein using the protein annotaiton
        from its two genes
        """

        fusion_domains = []
        gene5prime_domains = []
        gene3prime_domains = []

        tmp_domains = []

        # fetch the translation ids

        sqlite3_command = (
            "SELECT * FROM "
            + self.db_pointer.build
            + '_transcript WHERE transcript_stable_id=="'
            + self.transcript1.id
            + '"'
        )
        self.db_pointer.logger.debug("SQLite - " + sqlite3_command)
        self.db_pointer.sqlite3_cursor.execute(sqlite3_command)
        gene5prime_translation_id = self.db_pointer.sqlite3_cursor.fetchall()[0][3]

        sqlite3_command = (
            "SELECT * FROM "
            + self.db_pointer.build
            + '_transcript WHERE transcript_stable_id=="'
            + self.transcript2.id
            + '"'
        )
        self.db_pointer.logger.debug("SQLite - " + sqlite3_command)
        self.db_pointer.sqlite3_cursor.execute(sqlite3_command)
        gene3prime_translation_id = self.db_pointer.sqlite3_cursor.fetchall()[0][3]

        for protein_database in self.protein_databases:

            # fetch protein annotation

            sqlite3_command = (
                "SELECT * FROM "
                + self.db_pointer.build
                + "_"
                + protein_database
                + ' WHERE translation_id=="'
                + gene5prime_translation_id
                + '"'
            )
            self.db_pointer.logger.debug("SQLite - " + sqlite3_command)
            self.db_pointer.sqlite3_cursor.execute(sqlite3_command)
            tmp_domains += [list(x) for x in self.db_pointer.sqlite3_cursor.fetchall()]

        for d in tmp_domains:

            pfeature_id = d[2]
            pfeature_name = d[6]
            pfeature_description = d[5]
            pfeature_start = int(d[3])
            pfeature_end = int(d[4])

            gene5prime_domains.append(
                [
                    pfeature_id,
                    pfeature_name,
                    pfeature_description,
                    pfeature_start,
                    pfeature_end,
                ]
            )

            if self.transcript_protein_junction_5prime < pfeature_start:
                continue
            if self.transcript_protein_junction_5prime >= pfeature_end:

                fusion_domains.append(
                    [
                        pfeature_id,
                        pfeature_name,
                        pfeature_description,
                        pfeature_start,
                        pfeature_end,
                    ]
                )

            elif (
                self.transcript_protein_junction_5prime - pfeature_start
            ) >= utils.MIN_DOMAIN_LENGTH:

                pfeature_end = self.transcript_protein_junction_5prime

                fusion_domains.append(
                    [
                        pfeature_id,
                        pfeature_name,
                        pfeature_description,
                        pfeature_start,
                        pfeature_end,
                    ]
                )

        # only find the 3' gene partner's domains if the fusion is in-frame

        if self.effect != "out-of-frame":

            tmp_domains = []

            for database in self.protein_databases:

                sqlite3_command = (
                    "SELECT * FROM "
                    + self.db_pointer.build
                    + "_"
                    + database
                    + ' WHERE translation_id=="'
                    + gene3prime_translation_id
                    + '"'
                )
                self.db_pointer.logger.debug("SQLite - " + sqlite3_command)
                self.db_pointer.sqlite3_cursor.execute(sqlite3_command)
                tmp_domains += [list(x) for x in self.db_pointer.sqlite3_cursor.fetchall()]

            for d in tmp_domains:

                pfeature_ID = d[2]
                pfeature_name = d[6]
                pfeature_description = d[5]
                pfeature_start = int(d[3])
                pfeature_end = int(d[4])

                gene3prime_domains.append(
                    [
                        pfeature_ID,
                        pfeature_name,
                        pfeature_description,
                        pfeature_start,
                        pfeature_end,
                    ]
                )

                if self.transcript_protein_junction_3prime > pfeature_end:
                    continue
                if self.transcript_protein_junction_3prime <= pfeature_start:

                    pfeature_start = (
                        pfeature_start - self.transcript_protein_junction_3prime
                    ) + self.transcript_protein_junction_5prime
                    pfeature_end = (
                        pfeature_end - self.transcript_protein_junction_3prime
                    ) + self.transcript_protein_junction_5prime

                    fusion_domains.append(
                        [
                            pfeature_ID,
                            pfeature_name,
                            pfeature_description,
                            pfeature_start,
                            pfeature_end,
                        ]
                    )

                else:

                    pfeature_start = self.transcript_protein_junction_5prime
                    pfeature_end = (
                        pfeature_end - self.transcript_protein_junction_3prime
                    ) + self.transcript_protein_junction_5prime

                    fusion_domains.append(
                        [
                            pfeature_ID,
                            pfeature_name,
                            pfeature_description,
                            pfeature_start,
                            pfeature_end,
                        ]
                    )

        self.domains["fusion"] = fusion_domains
        self.domains[self.transcript1.id] = gene5prime_domains
        self.domains[self.transcript2.id] = gene3prime_domains

    def _fetch_protein(self):
        """
        Predict the potential protain amino acid sequence
        """

        self.transcript_protein_junction_5prime = 0
        self.transcript_protein_junction_3prime = 0

        self.protein_names = self.transcript1.protein_id + "-" + self.transcript2.protein_id

        self.transcript_protein_junction_5prime = int(self.transcript_cds_junction_5prime / 3.0)

        if (len(self.cds_5prime) / 3.0).is_integer() and (len(self.cds_3prime) / 3.0).is_integer():
            self.effect = "in-frame"
            self.transcript_protein_junction_3prime = int(self.transcript_cds_junction_3prime / 3.0)
        elif round((len(self.cds_5prime) / 3.0 % 1) + (len(self.cds_3prime) / 3.0 % 1), 2) == 1.0:
            self.effect = "in-frame (with mutation)"
            self.transcript_protein_junction_3prime = int(self.transcript_cds_junction_3prime / 3.0)
        else:
            self.effect = "out-of-frame"

        # check if CDS's length is multiple of 3, if not then print warning

        if (len(self.cds.seq) % 3) != 0:
            self.db_pointer.logger.warning("Length of fusion isoform CDS is not a multiple of 3!")

        # translate CDS into protein and remove everything after the stop codon

        if self.effect == "out-of-frame":

            # trim the CDS sequence if fusion is out-of-frame

            seq = self.cds.seq[0 : 3 * int(len(self.cds.seq) / 3)]

            protein_seq = seq.translate()
            protein_seq = protein_seq[0 : protein_seq.find("*")]
        else:
            protein_seq = self.cds.seq.translate()
            protein_seq = protein_seq[0 : protein_seq.find("*")]

        # predict molecular weight

        self.molecular_weight = SeqUtils.molecular_weight(protein_seq, seq_type="protein") / 1000.0

        # convert to a sequence record

        self.protein_length = len(protein_seq)

        self.protein = SeqRecord.SeqRecord(
            protein_seq,
            id=self.protein_names,
            name=self.protein_names,
            description=f"length: {self.protein_length}, "
            f"kD: {self.molecular_weight}, "
            f"transcripts: {self.name}, "
            f"strands: {self.transcript1.strand}/{self.transcript2.strand}, "
            f"genes: {self.gene_names}, effect: {self.effect}",
        )

    def _fetch_transcript_cds(self):
        """
        Predict the potential nucleotide sequence
        """

        self.transcript_cds_junction_5prime = 0
        self.transcript_cds_junction_3prime = 0

        # 5prime transcript

        if self.transcript1.strand == "+":
            for cds in self.transcript1.coding_sequence_position_ranges:
                if self.gene5prime.junction >= cds[1]:
                    self.transcript_cds_junction_5prime += cds[1] - cds[0] + 1
                elif self.gene5prime.junction <= cds[0]:
                    break
                else:
                    self.transcript_cds_junction_5prime += self.gene5prime.junction - cds[0] + 1
                    break
        else:
            for cds in self.transcript1.coding_sequence_position_ranges:
                if self.gene5prime.junction <= cds[0]:
                    self.transcript_cds_junction_5prime += cds[1] - cds[0] + 1
                elif self.gene5prime.junction >= cds[1]:
                    break
                else:
                    self.transcript_cds_junction_5prime += cds[1] - self.gene5prime.junction + 1

        self.cds_5prime = self.transcript1.coding_sequence[0 : self.transcript_cds_junction_5prime]

        if self.transcript2.strand == "+":
            for cds in self.transcript2.coding_sequence_position_ranges:
                if self.gene3prime.junction >= cds[1]:
                    self.transcript_cds_junction_3prime += cds[1] - cds[0] + 1
                elif self.gene3prime.junction <= cds[0]:
                    break
                else:
                    self.transcript_cds_junction_3prime += self.gene3prime.junction - cds[0]
        else:
            for cds in self.transcript2.coding_sequence_position_ranges:
                if self.gene3prime.junction <= cds[0]:
                    self.transcript_cds_junction_3prime += cds[1] - cds[0] + 1
                elif self.gene3prime.junction >= cds[1]:
                    break
                else:
                    self.transcript_cds_junction_3prime += cds[1] - self.gene3prime.junction

        self.cds_3prime = self.transcript2.coding_sequence[self.transcript_cds_junction_3prime : :]

        # create a sequence record

        if self.cds_5prime is not None and self.cds_3prime is not None:
            seq = self.cds_5prime + self.cds_3prime

        seq_length = len(self.cds_5prime + self.cds_3prime)

        self.cds = SeqRecord.SeqRecord(
            Seq.Seq(seq),
            id=self.name,
            name=self.name,
            description=f"length: {seq_length}, "
            f"genes: {self.transcript1.gene.name}/{self.transcript2.gene.name}, "
            f"strands: {self.transcript1.gene.strand}/{self.transcript2.gene.strand}",
        )

    def _fetch_transcript_cdna(self):
        """
        Predict the potential nucleotide sequence
        """

        self.transcript_cdna_junction_5prime = 0
        self.transcript_cdna_junction_3prime = 0

        # get the 5prime transcript sequence and determine if junction is
        # within intron

        n = 0

        n_max = len(self.transcript1.exon_intervals)
        exons = self.transcript1.exon_intervals

        if self.transcript1.strand == "+":

            exon_count = 0

            for exon in exons:

                exon_count += 1

                # is in intron?

                if n == 0 and self.gene5prime.junction < exon[0]:
                    self.effect_5prime = "intron (before cds)"

                if n == n_max and self.gene5prime.junction > exon[1]:
                    self.effect_5prime = "intron (after cds)"
                elif (
                    self.gene5prime.junction > exon[1]
                    and self.gene5prime.junction < exons[n + 1][0]
                ):
                    self.effect_5prime = "intron"

                n += 1

                # get sequence

                if self.gene5prime.junction >= exon[1]:
                    self.transcript_cdna_junction_5prime += exon[1] - exon[0] + 1
                    self.gene5prime_exon_intervals.append([exon[0], exon[1], exon_count])
                elif self.gene5prime.junction <= exon[0]:
                    break
                else:
                    self.transcript_cdna_junction_5prime += self.gene5prime.junction - exon[0] + 1
                    self.gene5prime_exon_intervals.append(
                        [exon[0], self.gene5prime.junction, exon_count]
                    )
                    break
        else:

            exon_count = 0

            for exon in exons:

                exon_count += 1

                # is in intron?

                if n == 0 and self.gene5prime.junction > exon[1]:
                    self.effect_5prime = "intron (before cds)"

                if n == n_max and self.gene5prime.junction < exon[0]:
                    self.effect_5prime = "intron (after cds)"
                elif (
                    self.gene5prime.junction < exon[0]
                    and self.gene5prime.junction > exons[n + 1][1]
                ):
                    self.effect_5prime = "intron"

                n += 1

                # get sequence

                if self.gene5prime.junction <= exon[0]:
                    self.transcript_cdna_junction_5prime += exon[1] - exon[0] + 1
                    self.gene5prime_exon_intervals.append([exon[0], exon[1], exon_count])
                elif self.gene5prime.junction >= exon[1]:
                    break
                else:
                    self.transcript_cdna_junction_5prime += exon[1] - self.gene5prime.junction + 1
                    self.gene5prime_exon_intervals.append(
                        [self.gene5prime.junction, exon[1], exon_count]
                    )

        try:
            self.cdna_5prime = self.transcript1.sequence[0 : self.transcript_cdna_junction_5prime]
        except TypeError:
            self.db_pointer.logger.warning(
                f"No cDNA sequence available for {self.gene5prime.gene.name}! "
                f"Will not print cDNA sequence for the {self.gene_names} fusion. "
                "You might be working with an outdated pyensembl. "
                "Update the package and rerun 'pyensembl install'"
            )

        # get the 3prime transcript sequence and determine if junction is
        # within intron

        n = 0
        n_max = len(self.transcript2.exon_intervals)
        exons = self.transcript2.exon_intervals

        if self.transcript2.strand == "+":

            exon_count = 0

            # get the exons in the fusion

            for exon in exons:
                exon_count += 1
                if self.gene3prime.junction >= exon[1]:
                    continue
                if self.gene3prime.junction <= exon[0]:
                    self.gene3prime_exon_intervals.append([exon[0], exon[1], exon_count])
                else:
                    self.gene3prime_exon_intervals.append(
                        [self.gene3prime.junction, exon[1], exon_count]
                    )

            for exon in exons:

                # is in intron?

                if n == 0 and self.gene3prime.junction < exon[0]:
                    self.effect_3prime = "intron (before cds)"

                if n == n_max and self.gene3prime.junction > exon[1]:
                    self.effect_3prime = "intron (after cds)"
                elif (
                    self.gene3prime.junction > exon[1]
                    and self.gene3prime.junction < exons[n + 1][0]
                ):
                    self.effect_3prime = "intron"

                n += 1

                # get sequence

                if self.gene3prime.junction >= exon[1]:
                    self.transcript_cdna_junction_3prime += exon[1] - exon[0] + 1
                elif self.gene3prime.junction <= exon[0]:
                    break
                else:
                    self.transcript_cdna_junction_3prime += self.gene3prime.junction - exon[0]
        else:

            exon_count = 0

            # get the exons in the fusion

            for exon in exons:
                exon_count += 1
                if self.gene3prime.junction <= exon[0]:
                    continue
                if self.gene3prime.junction >= exon[1]:
                    self.gene3prime_exon_intervals.append([exon[0], exon[1], exon_count])
                else:
                    self.gene3prime_exon_intervals.append(
                        [exon[0], self.gene3prime.junction, exon_count]
                    )

            for exon in exons:

                # is in intron?

                if n == 0 and self.gene3prime.junction > exon[1]:
                    self.effect_3prime = "intron (before cds)"

                if n == n_max and self.gene3prime.junction < exon[0]:
                    self.effect_3prime = "intron (after cds)"
                elif (
                    self.gene3prime.junction < exon[0]
                    and self.gene3prime.junction > exons[n + 1][1]
                ):
                    self.effect_3prime = "intron"

                n += 1

                # get sequence

                if self.gene3prime.junction <= exon[0]:
                    self.transcript_cdna_junction_3prime += exon[1] - exon[0] + 1
                elif self.gene3prime.junction >= exon[1]:
                    break
                else:
                    self.transcript_cdna_junction_3prime += exon[1] - self.gene3prime.junction

        if self.cdna_5prime is not None:
            try:
                self.cdna_3prime = self.transcript2.sequence[
                    self.transcript_cdna_junction_3prime : :
                ]
            except TypeError:
                self.db_pointer.logger.warning(
                    f"No cDNA sequence available for {self.gene3prime.gene.name}! "
                    f"Will not print cDNA sequence for the {self.gene_names} fusion. "
                    "You might be working with an outdated pyensembl. "
                    "Update the package and rerun 'pyensembl install'"
                )

        if self.cdna_5prime is not None and self.cdna_3prime is not None:
            seq = self.cdna_5prime + self.cdna_3prime
            seq_length = str(len(seq))
        else:
            seq = ""
            seq_length = "NA"

        self.cdna = SeqRecord.SeqRecord(
            Seq.Seq(seq),
            id=self.name,
            name=self.name,
            description="length=" + seq_length,
        )

        # find out if the junction on either gene is with in the 5' or 3' UTR,
        # or if it exactly at the beginning or end of the UTR

        # the 5' gene

        if self.effect_5prime.find("intron") == -1:

            if self.transcript1.complete and self.transcript_cdna_junction_5prime < len(
                self.transcript1.five_prime_utr_sequence
            ):
                self.effect_5prime = "5UTR"
            elif self.transcript1.complete and self.transcript_cdna_junction_5prime == len(
                self.transcript1.five_prime_utr_sequence
            ):
                self.effect_5prime = "5UTR (end)"

            if self.transcript1.complete and (
                len(self.transcript1) - self.transcript_cdna_junction_5prime
            ) < len(self.transcript1.three_prime_utr_sequence):
                self.effect_5prime = "3UTR"
            elif self.transcript1.complete and (
                len(self.transcript1) - self.transcript_cdna_junction_5prime
            ) == len(self.transcript1.three_prime_utr_sequence):
                self.effect_5prime = "3UTR (start)"

        # the 3' gene

        if self.effect_3prime.find("intron") == -1:

            if self.transcript2.complete and self.transcript_cdna_junction_3prime < len(
                self.transcript2.five_prime_utr_sequence
            ):
                self.effect_3prime = "5UTR"
            elif self.transcript2.complete and self.transcript_cdna_junction_3prime == len(
                self.transcript2.five_prime_utr_sequence
            ):
                self.effect_3prime = "5UTR (end)"

            if self.transcript2.complete and (
                len(self.transcript2) - self.transcript_cdna_junction_3prime
            ) < len(self.transcript2.three_prime_utr_sequence):
                self.effect_3prime = "3UTR"
            elif self.transcript2.complete and (
                len(self.transcript2) - self.transcript_cdna_junction_3prime
            ) == len(self.transcript2.three_prime_utr_sequence):
                self.effect_3prime = "3UTR (start)"

        self.effect = self.effect_5prime + "-" + self.effect_3prime

    # def _check_if_in_intron(self):

    def predict_effect(self):
        """
        For all gene isoform combinations predict the effect of the fusion
        (e.g. in-frame, out-of-frame, etc...). Then if it has coding potential
        then annotate with domains
        """

        # check if within exon or UTR

        self._fetch_transcript_cdna()

        # check if within CDS and if it occurs at the very beginning or end of CDS

        if self.transcript1.complete and (self.effect_5prime.find("UTR") == -1):

            if self.effect_5prime.find("intron") != -1:

                # if in intron, is it between CDS regions?

                n = 0
                n_max = len(self.transcript1.coding_sequence_position_ranges) - 1

                for cds in self.transcript1.coding_sequence_position_ranges:

                    if self.transcript1.strand == "+":
                        if n == 0 and self.gene5prime.junction < cds[0]:
                            self.effect_5prime = "intron (before cds)"
                            break
                        if n == n_max and self.gene5prime.junction > cds[1]:
                            self.effect_5prime = "intron (after cds)"
                            break
                    else:
                        if n == 0 and self.gene5prime.junction > cds[1]:
                            self.effect_5prime = "intron (before cds)"
                            break
                        if n == n_max and self.gene5prime.junction < cds[0]:
                            self.effect_5prime = "intron (after cds)"
                            break

                    if (
                        (n != 0)
                        and (n != n_max)
                        and (
                            self.gene5prime.junction < cds[0]
                            or self.gene5prime.junction
                            > self.transcript1.coding_sequence_position_ranges[n + 1][1]
                        )
                    ):

                        self.effect_5prime = "intron (cds)"
                        break

                    n += 1
            else:
                for cds in self.transcript1.coding_sequence_position_ranges:

                    if self.gene5prime.junction >= cds[0] and self.gene5prime.junction <= cds[1]:
                        self.effect_5prime = "CDS"
                        break

                if self.transcript1.strand == "+":
                    if (
                        self.gene5prime.junction
                        == self.transcript1.coding_sequence_position_ranges[0][0]
                    ):
                        self.effect_5prime = "CDS (start)"
                    elif (
                        self.gene5prime.junction
                        == self.transcript1.coding_sequence_position_ranges[-1][1]
                    ):
                        self.effect_5prime = "CDS (end)"
                else:
                    if (
                        self.gene5prime.junction
                        == self.transcript1.coding_sequence_position_ranges[0][1]
                    ):
                        self.effect_5prime = "CDS (start)"
                    elif (
                        self.gene5prime.junction
                        == self.transcript1.coding_sequence_position_ranges[-1][0]
                    ):
                        self.effect_5prime = "CDS (end)"

        if self.transcript2.complete and (self.effect_3prime.find("UTR") == -1):

            if self.effect_3prime.find("intron") != -1:

                # if in intron, is it between CDS regions?

                n = 0
                n_max = len(self.transcript2.coding_sequence_position_ranges) - 1

                for cds in self.transcript2.coding_sequence_position_ranges:

                    if self.transcript2.strand == "+":
                        if n == 0 and self.gene3prime.junction < cds[0]:
                            self.effect_3prime = "intron (before cds)"
                            break
                        if n == n_max and self.gene3prime.junction > cds[1]:
                            self.effect_3prime = "intron (after cds)"
                            break
                    else:
                        if n == 0 and self.gene3prime.junction > cds[1]:
                            self.effect_3prime = "intron (before cds)"
                            break
                        if n == n_max and self.gene3prime.junction < cds[0]:
                            self.effect_3prime = "intron (after cds)"
                            break

                    if (
                        (n != 0)
                        and (n != n_max)
                        and (
                            self.gene3prime.junction < cds[0]
                            or self.gene3prime.junction
                            > self.transcript2.coding_sequence_position_ranges[n + 1][1]
                        )
                    ):

                        self.effect_3prime = "intron (cds)"
                        break

                    n += 1

            else:
                for cds in self.transcript2.coding_sequence_position_ranges:
                    if self.gene3prime.junction >= cds[0] and self.gene3prime.junction <= cds[1]:
                        self.effect_3prime = "CDS"
                        break

                if self.transcript2.strand == "+":
                    if (
                        self.gene3prime.junction
                        == self.transcript2.coding_sequence_position_ranges[0][0]
                    ):
                        self.effect_3prime = "CDS (start)"
                    elif (
                        self.gene3prime.junction
                        == self.transcript2.coding_sequence_position_ranges[-1][1]
                    ):
                        self.effect_3prime = "CDS (end)"
                else:
                    if (
                        self.gene3prime.junction
                        == self.transcript2.coding_sequence_position_ranges[0][1]
                    ):
                        self.effect_3prime = "CDS (start)"
                    elif (
                        self.gene3prime.junction
                        == self.transcript2.coding_sequence_position_ranges[-1][0]
                    ):
                        self.effect_3prime = "CDS (end)"

        # get if has coding potential

        self.has_coding_potential = utils.CODING_COMBINATIONS[
            (self.effect_5prime, self.effect_3prime)
        ]["protein_coding_potential"]

        # check if they don't have stop and/or start codons

        reasons = []

        if not self.transcript1.contains_start_codon:
            self.has_coding_potential = False
            reasons.append("no known 5' transcript start codon")
        if not self.transcript1.contains_stop_codon:
            self.has_coding_potential = False
            reasons.append("no known 5' transcript stop codon")

        if not self.transcript2.contains_start_codon:
            self.has_coding_potential = False
            reasons.append("no known 3' transcript start codon")
        if not self.transcript2.contains_stop_codon:
            self.has_coding_potential = False
            reasons.append("no known 3' transcript stop codon")

        # append information to cdna fasta headers

        self.cdna.description += f"; locations: {self.effect_5prime}/{self.effect_3prime};"
        self.cdna.description += f" strands: {self.transcript1.strand}/{self.transcript2.strand};"
        self.cdna.description += f" Has protein coding potential: {self.has_coding_potential};"

        if not self.has_coding_potential:
            reasons = ", ".join(reasons)
            self.cdna.description += f" Reason: {reasons}"

        # if the fusion transcript has coding potential then
        # fetch its CDS and protein sequences and annotate it

        if self.has_coding_potential:
            self._fetch_transcript_cds()
            self._fetch_protein()
            self._annotate()



""" exceptions.py
Exceptions used for error handling.
"""


class DataBaseError(Exception):
    """_summary_

    Args:
        Exception (Exception): Base python exception.
    """

    def __init__(self, error):
        Exception.__init__(self, error)
        self.error = error


class GeneIDException(Exception):
    """Exception to handle case of no Ensemble gene ID for gene.

    Args:
        Exception (Exception): Base python exception.
    """

    def __init__(self, gene):
        Exception.__init__(
            self,
            f"No Ensembl ID found for {gene}! Check its spelling and if you are "
            "using the right genome build.",
        )


class TooManyGenesException(Exception):
    """Exception to handle case of too many genes.

    Args:
        Exception (Exception): Base python exception.
    """

    def __init__(self, gene, ids, build):
        ids = ", ".join(ids)
        Exception.__init__(
            self,
            f"Multiple Ensembl IDs found matching {gene}: {ids} for genome {build}."
            " Specify which Ensembl ID.",
        )


class JunctionException(Exception):
    """Exception to handle junction not being in the gene.

    Args:
        Exception (Exception): Base python exception.
    """

    def __init__(self, gene, junction):
        Exception.__init__(self, f"Junction {junction} not within {gene} gene boundaries!")



"""
All combinations except those that do not produce some sort of
chimeric protein are commented out
"""

import itertools
import os

agfusion_db = os.path.join(os.path.split(__file__)[0], "data", "agfusion.db")

AGFUSION_DB_URL = "https://s3.amazonaws.com/agfusion/agfusion."

# this is mostly contigent on the maximum ensembl release supported
# by pyensembl

MAX_ENSEMBL_RELEASE = 95

GENOME_SHORTCUTS = {
    "GRCm38": ["mus_musculus", MAX_ENSEMBL_RELEASE],
    "mm10": ["mus_musculus", MAX_ENSEMBL_RELEASE],
    "mm9": ["mus_musculus", 67],
    "GRCh38": ["homo_sapiens", MAX_ENSEMBL_RELEASE],
    "hg38": ["homo_sapiens", MAX_ENSEMBL_RELEASE],
    "hg19": ["homo_sapiens", 75],
}

AVAILABLE_ENSEMBL_SPECIES = {
    "homo_sapiens": range(69, MAX_ENSEMBL_RELEASE + 1),
    "mus_musculus": range(67, MAX_ENSEMBL_RELEASE + 1),
}

ENSEMBL_MYSQL_TABLES = {"homo_sapiens": {}, "mus_musculus": {}}

# ENSEMBL_MYSQL_TABLES['homo_sapiens'][48] = 'homo_sapiens_core_48_36j'
# ENSEMBL_MYSQL_TABLES['homo_sapiens'][49] = 'homo_sapiens_core_49_36k'
# ENSEMBL_MYSQL_TABLES['homo_sapiens'][50] = 'homo_sapiens_core_50_36l'
# ENSEMBL_MYSQL_TABLES['homo_sapiens'][51] = 'homo_sapiens_core_51_36m'
# ENSEMBL_MYSQL_TABLES['homo_sapiens'][52] = 'homo_sapiens_core_52_36n'
# ENSEMBL_MYSQL_TABLES['homo_sapiens'][53] = 'homo_sapiens_core_53_36o'
# ENSEMBL_MYSQL_TABLES['homo_sapiens'][54] = 'homo_sapiens_core_54_36p'
# ENSEMBL_MYSQL_TABLES['homo_sapiens'][55] = 'homo_sapiens_core_55_37'
# ENSEMBL_MYSQL_TABLES['homo_sapiens'][56] = 'homo_sapiens_core_56_37a'
# ENSEMBL_MYSQL_TABLES['homo_sapiens'][57] = 'homo_sapiens_core_57_37b'
# ENSEMBL_MYSQL_TABLES['homo_sapiens'][58] = 'homo_sapiens_core_58_37c'
# ENSEMBL_MYSQL_TABLES['homo_sapiens'][59] = 'homo_sapiens_core_59_37d'
# ENSEMBL_MYSQL_TABLES['homo_sapiens'][60] = 'homo_sapiens_core_60_37e'
# ENSEMBL_MYSQL_TABLES['homo_sapiens'][61] = 'homo_sapiens_core_61_37f'
# ENSEMBL_MYSQL_TABLES['homo_sapiens'][62] = 'homo_sapiens_core_62_37g'

for i in range(63, MAX_ENSEMBL_RELEASE + 1):
    if i < 76:
        ENSEMBL_MYSQL_TABLES["homo_sapiens"][i] = "homo_sapiens_core_" + str(i) + "_37"
    else:
        ENSEMBL_MYSQL_TABLES["homo_sapiens"][i] = "homo_sapiens_core_" + str(i) + "_38"

for i in range(67, MAX_ENSEMBL_RELEASE + 1):
    if i < 68:
        ENSEMBL_MYSQL_TABLES["mus_musculus"][i] = "mus_musculus_core_" + str(i) + "_37"
    else:
        ENSEMBL_MYSQL_TABLES["mus_musculus"][i] = "mus_musculus_core_" + str(i) + "_38"

# min amino acid length of domain to plot it

MIN_DOMAIN_LENGTH = 5

# just

STANDARD_CHROMOSOMES = [str(i) for i in range(1, 23)] + ["X", "Y", "MT"]

# the available protein domain annotations

PROTEIN_ANNOTATIONS = [
    "pfam",
    "smart",
    "superfamily",
    "tigrfam",
    "pfscan",
    "tmhmm",
    "seg",
    "ncoils",
    "prints",
    "pirsf",
    "signalp",
]

JUNCTION_LOCATIONS = [
    "CDS",
    "CDS (start)",
    "CDS (end)",
    "5UTR",
    "5UTR (end)",
    "3UTR",
    "3UTR (start)",
    "exon",
    "intron",
    "intron (cds)",
    "intron (before cds)",
    "intron (after cds)",
]

CODING_COMBINATIONS = list(itertools.product(JUNCTION_LOCATIONS, JUNCTION_LOCATIONS))
CODING_COMBINATIONS = {
    i: {"protein_coding_potential": False, "truncated": False} for i in CODING_COMBINATIONS
}


CODING_COMBINATIONS["CDS", "CDS"] = {
    "protein_coding_potential": True,
    "truncated": False,
}
CODING_COMBINATIONS["CDS", "CDS (start)"] = {
    "protein_coding_potential": True,
    "truncated": False,
}
CODING_COMBINATIONS["CDS", "CDS (end)"] = {
    "protein_coding_potential": True,
    "truncated": True,
}
CODING_COMBINATIONS["CDS", "5UTR"] = {
    "protein_coding_potential": True,
    "truncated": True,
}
CODING_COMBINATIONS["CDS", "5UTR (end)"] = {
    "protein_coding_potential": True,
    "truncated": True,
}
CODING_COMBINATIONS["CDS", "3UTR"] = {
    "protein_coding_potential": True,
    "truncated": True,
}
CODING_COMBINATIONS["CDS", "3UTR (start)"] = {
    "protein_coding_potential": True,
    "truncated": True,
}
CODING_COMBINATIONS["CDS", "intron"] = {
    "protein_coding_potential": True,
    "truncated": True,
}
CODING_COMBINATIONS["CDS", "intron (cds)"] = {
    "protein_coding_potential": True,
    "truncated": False,
}
CODING_COMBINATIONS["CDS", "intron (before cds)"] = {
    "protein_coding_potential": True,
    "truncated": True,
}
CODING_COMBINATIONS["CDS", "intron (after cds)"] = {
    "protein_coding_potential": True,
    "truncated": True,
}

CODING_COMBINATIONS["CDS (end)", "CDS"] = {
    "protein_coding_potential": True,
    "truncated": False,
}
CODING_COMBINATIONS["CDS (end)", "CDS (start)"] = {
    "protein_coding_potential": True,
    "truncated": False,
}
CODING_COMBINATIONS["CDS (end)", "CDS (end)"] = {
    "protein_coding_potential": True,
    "truncated": False,
}
CODING_COMBINATIONS["CDS (end)", "5UTR"] = {
    "protein_coding_potential": True,
    "truncated": False,
}
CODING_COMBINATIONS["CDS (end)", "5UTR (end)"] = {
    "protein_coding_potential": True,
    "truncated": False,
}
CODING_COMBINATIONS["CDS (end)", "3UTR"] = {
    "protein_coding_potential": True,
    "truncated": False,
}
CODING_COMBINATIONS["CDS (end)", "3UTR (start)"] = {
    "protein_coding_potential": True,
    "truncated": False,
}
CODING_COMBINATIONS["CDS (end)", "intron"] = {
    "protein_coding_potential": True,
    "truncated": False,
}
CODING_COMBINATIONS["CDS (end)", "intron (cds)"] = {
    "protein_coding_potential": True,
    "truncated": False,
}
CODING_COMBINATIONS["CDS (end)", "intron (before cds)"] = {
    "protein_coding_potential": True,
    "truncated": False,
}
CODING_COMBINATIONS["CDS (end)", "intron (after cds)"] = {
    "protein_coding_potential": True,
    "truncated": False,
}

# CODING_COMBINATIONS['CDS (start)','CDS'] = {'protein_coding_potential':True,'truncated':False}
CODING_COMBINATIONS["CDS (start)", "CDS (start)"] = {
    "protein_coding_potential": True,
    "truncated": False,
}
# CODING_COMBINATIONS['CDS (start)','CDS (end)'] = {
# 'protein_coding_potential':True,'truncated':False}
# CODING_COMBINATIONS['CDS (start)','5UTR'] = {
# 'protein_coding_potential':True,'truncated':False}
# CODING_COMBINATIONS['CDS (start)','5UTR (end)'] = {
# 'protein_coding_potential':True,'truncated':False}
# CODING_COMBINATIONS['CDS (start)','3UTR'] = {
# 'protein_coding_potential':True,'truncated':False}
# CODING_COMBINATIONS['CDS (start)','3UTR (start)'] = {
# 'protein_coding_potential':True,'truncated':False}
CODING_COMBINATIONS["CDS (start)", "intron"] = {
    "protein_coding_potential": True,
    "truncated": True,
}
CODING_COMBINATIONS["CDS (start)", "intron (cds)"] = {
    "protein_coding_potential": True,
    "truncated": False,
}
CODING_COMBINATIONS["CDS (start)", "intron (before cds)"] = {
    "protein_coding_potential": True,
    "truncated": True,
}
CODING_COMBINATIONS["CDS (start)", "intron (after cds)"] = {
    "protein_coding_potential": True,
    "truncated": True,
}

# CODING_COMBINATIONS['5UTR','CDS'] = {'protein_coding_potential':True,'truncated':False}
CODING_COMBINATIONS["5UTR", "CDS (start)"] = {
    "protein_coding_potential": True,
    "truncated": False,
}
# CODING_COMBINATIONS['5UTR','CDS (end)'] = {'protein_coding_potential':True,'truncated':False}
# CODING_COMBINATIONS['5UTR','5UTR'] = {'protein_coding_potential':True,'truncated':False}
# CODING_COMBINATIONS['5UTR','5UTR (end)'] = {'protein_coding_potential':True,'truncated':False}
# CODING_COMBINATIONS['5UTR','3UTR'] = {'protein_coding_potential':True,'truncated':False}
# CODING_COMBINATIONS['5UTR','3UTR (start)'] = {'protein_coding_potential':True,'truncated':False}
# CODING_COMBINATIONS['5UTR','intron'] = {'protein_coding_potential':True,'truncated':False}
# CODING_COMBINATIONS['5UTR','intron (cds)'] = {'protein_coding_potential':True,'truncated':False}
CODING_COMBINATIONS["5UTR", "intron (before cds)"] = {
    "protein_coding_potential": True,
    "truncated": False,
}
# CODING_COMBINATIONS['5UTR','intron (after cds)'] = {
# 'protein_coding_potential':True,'truncated':False}

# CODING_COMBINATIONS['5UTR (end)','CDS'] = {'protein_coding_potential':True,'truncated':False}
CODING_COMBINATIONS["5UTR (end)", "CDS (start)"] = {
    "protein_coding_potential": True,
    "truncated": False,
}
# CODING_COMBINATIONS['5UTR (end)','CDS (end)'] = {
# 'protein_coding_potential':True,'truncated':False}
# CODING_COMBINATIONS['5UTR (end)','5UTR'] = {
# 'protein_coding_potential':True,'truncated':False}
# CODING_COMBINATIONS['5UTR (end)','5UTR (end)'] = {
# 'protein_coding_potential':True,'truncated':False}
# CODING_COMBINATIONS['5UTR (end)','3UTR'] = {
# 'protein_coding_potential':True,'truncated':False}
# CODING_COMBINATIONS['5UTR (end)','3UTR (start)'] = {
# 'protein_coding_potential':True,'truncated':False}
# CODING_COMBINATIONS['5UTR (end)','intron (cds)'] = {
# 'protein_coding_potential':True,'truncated':False}
CODING_COMBINATIONS["5UTR (end)", "intron (before cds)"] = {
    "protein_coding_potential": True,
    "truncated": False,
}
# CODING_COMBINATIONS['5UTR (end)','intron (after cds)'] = {
# 'protein_coding_potential':True,'truncated':False}

CODING_COMBINATIONS["3UTR", "CDS"] = {
    "protein_coding_potential": True,
    "truncated": False,
}
CODING_COMBINATIONS["3UTR", "CDS (start)"] = {
    "protein_coding_potential": True,
    "truncated": False,
}
CODING_COMBINATIONS["3UTR", "CDS (end)"] = {
    "protein_coding_potential": True,
    "truncated": False,
}
CODING_COMBINATIONS["3UTR", "5UTR"] = {
    "protein_coding_potential": True,
    "truncated": False,
}
CODING_COMBINATIONS["3UTR", "5UTR (end)"] = {
    "protein_coding_potential": True,
    "truncated": False,
}
CODING_COMBINATIONS["3UTR", "3UTR"] = {
    "protein_coding_potential": True,
    "truncated": False,
}
CODING_COMBINATIONS["3UTR", "3UTR (start)"] = {
    "protein_coding_potential": True,
    "truncated": False,
}
CODING_COMBINATIONS["3UTR", "intron"] = {
    "protein_coding_potential": True,
    "truncated": False,
}
CODING_COMBINATIONS["3UTR", "intron (cds)"] = {
    "protein_coding_potential": True,
    "truncated": False,
}
CODING_COMBINATIONS["3UTR", "intron (before cds)"] = {
    "protein_coding_potential": True,
    "truncated": False,
}
CODING_COMBINATIONS["3UTR", "intron (after cds)"] = {
    "protein_coding_potential": True,
    "truncated": False,
}

CODING_COMBINATIONS["3UTR (start)", "CDS"] = {
    "protein_coding_potential": True,
    "truncated": False,
}
CODING_COMBINATIONS["3UTR (start)", "CDS (start)"] = {
    "protein_coding_potential": True,
    "truncated": False,
}
CODING_COMBINATIONS["3UTR (start)", "CDS (end)"] = {
    "protein_coding_potential": True,
    "truncated": False,
}
CODING_COMBINATIONS["3UTR (start)", "5UTR"] = {
    "protein_coding_potential": True,
    "truncated": False,
}
CODING_COMBINATIONS["3UTR (start)", "5UTR (end)"] = {
    "protein_coding_potential": True,
    "truncated": False,
}
CODING_COMBINATIONS["3UTR (start)", "3UTR"] = {
    "protein_coding_potential": True,
    "truncated": False,
}
CODING_COMBINATIONS["3UTR (start)", "3UTR (start)"] = {
    "protein_coding_potential": True,
    "truncated": False,
}
CODING_COMBINATIONS["3UTR (start)", "intron"] = {
    "protein_coding_potential": True,
    "truncated": False,
}
CODING_COMBINATIONS["3UTR (start)", "intron (cds)"] = {
    "protein_coding_potential": True,
    "truncated": False,
}
CODING_COMBINATIONS["3UTR (start)", "intron (after cds)"] = {
    "protein_coding_potential": True,
    "truncated": False,
}
CODING_COMBINATIONS["3UTR (start)", "intron (before cds)"] = {
    "protein_coding_potential": True,
    "truncated": False,
}

# CODING_COMBINATIONS['3UTR (end)','CDS'] = {
# 'protein_coding_potential':True,'truncated':False}
# CODING_COMBINATIONS['3UTR (end)','CDS (start)'] = {
# 'protein_coding_potential':True,'truncated':False}
# CODING_COMBINATIONS['3UTR (end)','CDS (end)'] = {
# 'protein_coding_potential':True,'truncated':False}
# CODING_COMBINATIONS['3UTR (end)','5UTR'] = {
# 'protein_coding_potential':True,'truncated':False}
# CODING_COMBINATIONS['3UTR (end)','5UTR (end)'] = {
# 'protein_coding_potential':True,'truncated':False}
# CODING_COMBINATIONS['3UTR (end)','3UTR'] = {
# 'protein_coding_potential':True,'truncated':False}
# CODING_COMBINATIONS['3UTR (end)','3UTR (start)'] = {
# 'protein_coding_potential':True,'truncated':False}
# CODING_COMBINATIONS['3UTR (end)','intron'] = {
# 'protein_coding_potential':True,'truncated':False}
# CODING_COMBINATIONS['3UTR (end)','intron (cds)'] = {
# 'protein_coding_potential':True,'truncated':False}
CODING_COMBINATIONS["3UTR (end)", "intron (before cds)"] = {
    "protein_coding_potential": True,
    "truncated": False,
}
# CODING_COMBINATIONS['3UTR (end)','intron (after cds)'] = {
# 'protein_coding_potential':True,'truncated':False}

# CODING_COMBINATIONS['intron','CDS'] = {'protein_coding_potential':True,'truncated':False}
CODING_COMBINATIONS["intron", "CDS (start)"] = {
    "protein_coding_potential": True,
    "truncated": False,
}
# CODING_COMBINATIONS['intron','CDS (end)'] = {'protein_coding_potential':True,'truncated':False}
CODING_COMBINATIONS["intron", "5UTR"] = {
    "protein_coding_potential": True,
    "truncated": False,
}
CODING_COMBINATIONS["intron", "5UTR (end)"] = {
    "protein_coding_potential": True,
    "truncated": False,
}
# CODING_COMBINATIONS['intron','3UTR'] = {'protein_coding_potential':True,'truncated':False}
# CODING_COMBINATIONS['intron','3UTR (start)'] = {'protein_coding_potential':True,'truncated':False}
# CODING_COMBINATIONS['intron','intron'] = {'protein_coding_potential':True,'truncated':False}
# CODING_COMBINATIONS['intron','intron (cds)'] = {'protein_coding_potential':True,'truncated':False}
CODING_COMBINATIONS["intron", "intron (before cds)"] = {
    "protein_coding_potential": True,
    "truncated": False,
}
# CODING_COMBINATIONS['intron','intron (after cds)'] = {
#     'protein_coding_potential':{
#         'protein_coding_potential':True,
#         'truncated':False
#     },
#     'truncated':False
# }

CODING_COMBINATIONS["intron (cds)", "CDS"] = {
    "protein_coding_potential": True,
    "truncated": False,
}
CODING_COMBINATIONS["intron (cds)", "CDS (start)"] = {
    "protein_coding_potential": True,
    "truncated": False,
}
CODING_COMBINATIONS["intron (cds)", "CDS (end)"] = {
    "protein_coding_potential": True,
    "truncated": True,
}
CODING_COMBINATIONS["intron (cds)", "5UTR"] = {
    "protein_coding_potential": True,
    "truncated": True,
}
CODING_COMBINATIONS["intron (cds)", "5UTR (end)"] = {
    "protein_coding_potential": True,
    "truncated": True,
}
CODING_COMBINATIONS["intron (cds)", "3UTR"] = {
    "protein_coding_potential": True,
    "truncated": True,
}
CODING_COMBINATIONS["intron (cds)", "3UTR (start)"] = {
    "protein_coding_potential": True,
    "truncated": True,
}
CODING_COMBINATIONS["intron (cds)", "intron"] = {
    "protein_coding_potential": True,
    "truncated": True,
}
CODING_COMBINATIONS["intron (cds)", "intron (cds)"] = {
    "protein_coding_potential": True,
    "truncated": False,
}
# CODING_COMBINATIONS['intron (cds)','intron (before cds)'] = {
# 'protein_coding_potential':True,'truncated':False}
# CODING_COMBINATIONS['intron (cds)','intron (after cds)'] = {
# 'protein_coding_potential':True,'truncated':False}

# CODING_COMBINATIONS['intron (before cds)','CDS'] = {
# 'protein_coding_potential':True,'truncated':False}
CODING_COMBINATIONS["intron (before cds)", "CDS (start)"] = {
    "protein_coding_potential": True,
    "truncated": False,
}
# CODING_COMBINATIONS['intron (before cds)','CDS (end)'] = {
# 'protein_coding_potential':True,'truncated':False}
CODING_COMBINATIONS["intron (before cds)", "5UTR"] = {
    "protein_coding_potential": True,
    "truncated": False,
}
CODING_COMBINATIONS["intron (before cds)", "5UTR (end)"] = {
    "protein_coding_potential": True,
    "truncated": False,
}
# CODING_COMBINATIONS['intron (before cds)','3UTR'] = {
# 'protein_coding_potential':True,'truncated':False}
# CODING_COMBINATIONS['intron (before cds)','3UTR (start)'] = {
# 'protein_coding_potential':True,'truncated':False}
# CODING_COMBINATIONS['intron (before cds)','intron'] = {
# 'protein_coding_potential':True,'truncated':False}
# CODING_COMBINATIONS['intron (before cds)','intron (cds)'] = {
# 'protein_coding_potential':True,'truncated':False}
CODING_COMBINATIONS["intron (before cds)", "intron (before cds)"] = {
    "protein_coding_potential": True,
    "truncated": False,
}
# CODING_COMBINATIONS['intron (before cds)','intron (after cds)'] = {
# 'protein_coding_potential':True,'truncated':False}

CODING_COMBINATIONS["intron (after cds)", "CDS"] = {
    "protein_coding_potential": True,
    "truncated": False,
}
CODING_COMBINATIONS["intron (after cds)", "CDS (start)"] = {
    "protein_coding_potential": True,
    "truncated": False,
}
CODING_COMBINATIONS["intron (after cds)", "CDS (end)"] = {
    "protein_coding_potential": True,
    "truncated": False,
}
CODING_COMBINATIONS["intron (after cds)", "5UTR"] = {
    "protein_coding_potential": True,
    "truncated": False,
}
CODING_COMBINATIONS["intron (after cds)", "5UTR (end)"] = {
    "protein_coding_potential": True,
    "truncated": False,
}
CODING_COMBINATIONS["intron (after cds)", "3UTR"] = {
    "protein_coding_potential": True,
    "truncated": False,
}
CODING_COMBINATIONS["intron (after cds)", "3UTR (start)"] = {
    "protein_coding_potential": True,
    "truncated": False,
}
CODING_COMBINATIONS["intron (after cds)", "intron"] = {
    "protein_coding_potential": True,
    "truncated": False,
}
CODING_COMBINATIONS["intron (after cds)", "intron (cds)"] = {
    "protein_coding_potential": True,
    "truncated": False,
}
CODING_COMBINATIONS["intron (after cds)", "intron (before cds)"] = {
    "protein_coding_potential": True,
    "truncated": False,
}
CODING_COMBINATIONS["intron (after cds)", "intron (after cds)"] = {
    "protein_coding_potential": True,
    "truncated": False,
}



"""Unit tests.
"""
import unittest
from os.path import abspath, curdir, join

import pyensembl

from agfusion import database, model, parsers

data_mouse = pyensembl.EnsemblRelease(84, "mouse")
db_mouse = database.AGFusionDB(abspath(join(curdir, "agfusion.mus_musculus.84.db")))
db_mouse.build = "mus_musculus_84"

data_human = pyensembl.EnsemblRelease(75, "human")
db_human = database.AGFusionDB(abspath(join(curdir, "agfusion.homo_sapiens.75.db")))
db_human.build = "homo_sapiens_75"

data_human95 = pyensembl.EnsemblRelease(95, "human")
db_human95 = database.AGFusionDB(abspath(join(curdir, "agfusion.homo_sapiens.95.db")))
db_human95.build = "homo_sapiens_95"


BASEDIR = "./data/FusionsFindingAlgorithms"


class TestFusionCatcher(unittest.TestCase):
    """Test parse FusionCatcher parse."""

    def test_parse(self):
        """Test basic parsing."""

        all_fusions = [
            "Adamts9_Ano2",
            "Trp53_Sat2",
            "1700112E06Rik_Runx1",
            "Runx1_1700112E06Rik",
            "Rell1_Lhfpl3",
            "Phc1_Smarca2",
            "Lrrc8d_Gbp11",
            "C920009B18Rik_H60b",
        ]
        for fusion in parsers.parsers["fusioncatcher"](
            f"{BASEDIR}/FusionCatcher/final-list_candidate-fusion-genes.txt",
            db_mouse.logger,
        ):
            fusion = model.Fusion(
                gene5prime=fusion["gene5prime"],
                gene5primejunction=fusion["gene5prime_junction"],
                gene3prime=fusion["gene3prime"],
                gene3primejunction=fusion["gene3prime_junction"],
                db=db_mouse,
                pyensembl_data=data_mouse,
                protein_databases=["pfam"],
                noncanonical=False,
            )
            assert fusion.name in all_fusions, f"{fusion.name} not in list!"


class TestArriba(unittest.TestCase):
    """Test parse Arriba parse."""

    def test_parse(self):
        """Test basic parsing."""

        all_fusions = [
            "BCR_ABL1",
        ]
        for fusion in parsers.parsers["arriba"](
            f"{BASEDIR}/Arriba/fusions.tsv",
            db_human.logger,
        ):
            fusion = model.Fusion(
                gene5prime=fusion["gene5prime"],
                gene5primejunction=fusion["gene5prime_junction"],
                gene3prime=fusion["gene3prime"],
                gene3primejunction=fusion["gene3prime_junction"],
                db=db_human,
                pyensembl_data=data_human,
                protein_databases=["pfam"],
                noncanonical=False,
            )
            assert fusion.name in all_fusions, f"{fusion.name} not in list!"


class TestSTARFusion(unittest.TestCase):
    """Test parse STAR Fusion"""

    def test_basic(self):
        """Test basic parsing."""

        all_fusions = ["ACACA_STAC2", "RPS6KB1_SNF8"]
        for fusion in parsers.parsers["starfusion"](
            f"{BASEDIR}/STARFusion/" + "star-fusion.fusion_candidates.final.abridged",
            db_human.logger,
        ):
            fusion = model.Fusion(
                gene5prime=fusion["gene5prime"],
                gene5primejunction=fusion["gene5prime_junction"],
                gene3prime=fusion["gene3prime"],
                gene3primejunction=fusion["gene3prime_junction"],
                db=db_human,
                pyensembl_data=data_human,
                protein_databases=["pfam"],
                noncanonical=False,
            )
            assert fusion.name in all_fusions, f"{fusion.name} not in list!"

    def test_with_coding_effect(self):
        """Test parse output with coding effect."""

        all_fusions = ["ARID3B_MYCNUT", "ARID3B_MYCN", "TVP23C_CDRT4"]
        for fusion in parsers.parsers["starfusion"](
            f"{BASEDIR}/STARFusion/" + "star-fusion.fusion_predictions.abridged.coding_effect.tsv",
            db_human95.logger,
        ):
            fusion = model.Fusion(
                gene5prime=fusion["gene5prime"],
                gene5primejunction=fusion["gene5prime_junction"],
                gene3prime=fusion["gene3prime"],
                gene3primejunction=fusion["gene3prime_junction"],
                db=db_human95,
                pyensembl_data=data_human95,
                protein_databases=["pfam"],
                noncanonical=False,
            )
            assert fusion.name in all_fusions, f"{fusion.name} not in list!"


class TestLongGf(unittest.TestCase):
    """Test parse LongGF"""

    def test_parse_mouse(self):
        """Test basic parsing."""

        all_fusions = ["Mocos_Rprd1a", "Ubc_Ubb", "Ubc_Gm11808", "Gm21887_Gm47283"]
        for fusion in parsers.parsers["longgf"](
            f"{BASEDIR}/LongGF/fusions_mouse.log",
            db_mouse.logger,
        ):
            fusion = model.Fusion(
                gene5prime=fusion["gene5prime"],
                gene5primejunction=fusion["gene5prime_junction"],
                gene3prime=fusion["gene3prime"],
                gene3primejunction=fusion["gene3prime_junction"],
                db=db_mouse,
                pyensembl_data=data_mouse,
                protein_databases=["pfam"],
                noncanonical=False,
            )
            assert fusion.name in all_fusions, f"{fusion.name} not in list!"

    def test_parse_human(self):
        """Test basic parsing."""

        all_fusions = ["BCAS4_BCAS3", "HNRNPC_ACIN1"]
        for fusion in parsers.parsers["longgf"](
            f"{BASEDIR}/LongGF/fusions_hg38.log",
            db_human95.logger,
        ):
            fusion = model.Fusion(
                gene5prime=fusion["gene5prime"],
                gene5primejunction=fusion["gene5prime_junction"],
                gene3prime=fusion["gene3prime"],
                gene3primejunction=fusion["gene3prime_junction"],
                db=db_human95,
                pyensembl_data=data_human95,
                protein_databases=["pfam"],
                noncanonical=False,
            )
            assert fusion.name in all_fusions, f"{fusion.name} not in list!"


class TestFusionInspector(unittest.TestCase):
    """Test parse FusionInspector"""

    def test_parse_human(self):
        """Test basic parsing."""

        all_fusions = ["AL627171.2_TPM3", "STAT3_AL627171.2"]

        for fusion in parsers.parsers["fusioninspector"](
            f"{BASEDIR}/FusionInspector/test.FusionInspector.fusions.abridged.txt",
            db_human95.logger,
        ):
            fusion = model.Fusion(
                gene5prime=fusion["gene5prime"],
                gene5primejunction=fusion["gene5prime_junction"],
                gene3prime=fusion["gene3prime"],
                gene3primejunction=fusion["gene3prime_junction"],
                db=db_human95,
                pyensembl_data=data_human95,
                protein_databases=["pfam"],
                noncanonical=False,
            )
            assert fusion.name in all_fusions, f"{fusion.name} not in list!"

        for fusion in parsers.parsers["fusioninspector"](
            f"{BASEDIR}/FusionInspector/test.FusionInspector.fusions.txt",
            db_human95.logger,
        ):
            fusion = model.Fusion(
                gene5prime=fusion["gene5prime"],
                gene5primejunction=fusion["gene5prime_junction"],
                gene3prime=fusion["gene3prime"],
                gene3primejunction=fusion["gene3prime_junction"],
                db=db_human95,
                pyensembl_data=data_human95,
                protein_databases=["pfam"],
                noncanonical=False,
            )
            assert fusion.name in all_fusions, f"{fusion.name} not in list!"


if __name__ == "__main__":
    unittest.main()



"""Unit tests.
"""
import unittest
from os.path import abspath, curdir, join

import pyensembl
from Bio import Seq, SeqIO

from agfusion import database, model, parsers

data = pyensembl.EnsemblRelease(84, "mouse")
db = database.AGFusionDB(abspath(join(curdir, "agfusion.mus_musculus.84.db")))
db.build = "mus_musculus_84"

data_human = pyensembl.EnsemblRelease(75, "human")
db_human = database.AGFusionDB(abspath(join(curdir, "agfusion.homo_sapiens.75.db")))
db_human.build = "homo_sapiens_75"


class TestSequencePredictionHuman(unittest.TestCase):
    """Test correctly predict human fusions"""

    def test_1(self):
        """
        test CDS and protein correct for junction that is on exon boundaries and
        produces an out-of-frame protein.
        """

        # test the dna and protein coding sequences are correct by comparing
        # with manually generally sequences

        fusion = model.Fusion(
            gene5prime="TMEM87B",
            gene5primejunction=112843681,
            gene3prime="MERTK",
            gene3primejunction=112722768,
            db=db_human,
            pyensembl_data=data_human,
            protein_databases=["pfam", "tmhmm"],
            noncanonical=False,
        )

        fusion.save_transcript_cdna("TMEM87B-MERTK-case0")
        fusion.save_transcript_cds("TMEM87B-MERTK-case0")
        fusion.save_proteins("TMEM87B-MERTK-case0")
        # fusion.save_images('DLG1-BRAF_mouse')

        test_cds = open("./data/test-human-case-0.txt", "r").read().strip()
        test_protein = Seq.Seq(test_cds).translate()
        test_protein = test_protein[0 : test_protein.find("*")]

        trans = fusion.transcripts["ENST00000283206_ENST00000295408"]

        assert test_cds == trans.cds.seq, "cds is wrongly predicted for human fusion (case 0)"
        assert (
            test_protein == trans.protein.seq
        ), "protein is wrongly predicted for human fusion (case 0)"

    def test_2(self):
        """
        test the dna and protein coding sequences are correct by comparing
        with manually generally sequences
        """

        fusion = model.Fusion(
            gene5prime="TMEM87B",
            gene5primejunction=112843681,
            gene3prime="MERTK",
            gene3primejunction=112722769,
            db=db_human,
            pyensembl_data=data_human,
            protein_databases=["pfam", "tmhmm"],
            noncanonical=False,
        )

        fusion.save_transcript_cdna("TMEM87B-MERTK-case2")
        fusion.save_transcript_cds("TMEM87B-MERTK-case2")
        fusion.save_proteins("TMEM87B-MERTK-case2")
        # fusion.save_images('DLG1-BRAF_mouse')

        test_cds = open("./data/test-human-case-2.txt", "r").read().strip()
        test_protein = Seq.Seq(test_cds).translate()
        test_protein = test_protein[0 : test_protein.find("*")]

        trans = fusion.transcripts["ENST00000283206_ENST00000295408"]

        assert test_cds == trans.cds.seq, "cds is wrongly predicted for human fusion (case 2)"
        assert (
            test_protein == trans.protein.seq
        ), "protein is wrongly predicted for human fusion (case 2)"

    def test_3(self):
        """
        test the dna and protein coding sequences are correct by comparing
        with manually generally sequences
        """

        fusion = model.Fusion(
            gene5prime="TMEM87B",
            gene5primejunction=112843681,
            gene3prime="MERTK",
            gene3primejunction=112722771,
            db=db_human,
            pyensembl_data=data_human,
            protein_databases=["pfam", "tmhmm"],
            noncanonical=False,
        )

        fusion.save_transcript_cdna("TMEM87B-MERTK-case3")
        fusion.save_transcript_cds("TMEM87B-MERTK-case3")
        fusion.save_proteins("TMEM87B-MERTK-case3")
        # fusion.save_images('DLG1-BRAF_mouse')

        test_cds = open("./data/test-human-case-3.txt", "r").read().strip()
        test_protein = Seq.Seq(test_cds).translate()
        test_protein = test_protein[0 : test_protein.find("*")]

        trans = fusion.transcripts["ENST00000283206_ENST00000295408"]

        assert test_cds == trans.cds.seq, "cds is wrongly predicted for human fusion (case 3)"
        assert (
            test_protein == trans.protein.seq
        ), "protein is wrongly predicted for human fusion (case 3)"


class TestSequencePrediction(unittest.TestCase):
    """Test correctly predict sequence content."""

    def test_1(self):
        """
        test CDS and cDNA correct for junction that is on exon boundaries and
        produces an in-frame protein.
        """

        # test the dna and protein coding sequences are correct by comparing
        # with manually generally sequences

        fusion = model.Fusion(
            gene5prime="ENSMUSG00000022770",
            gene5primejunction=31684294,
            gene3prime="ENSMUSG00000002413",
            gene3primejunction=39648486,
            db=db,
            pyensembl_data=data,
            protein_databases=["pfam", "tmhmm"],
            noncanonical=True,
        )

        fusion.save_transcript_cdna("DLG1-BRAF_mouse")
        fusion.save_transcript_cds("DLG1-BRAF_mouse")
        fusion.save_proteins("DLG1-BRAF_mouse")
        # fusion.save_images('DLG1-BRAF_mouse')

        test_cdna = SeqIO.parse(open("./data/Dlg1-Braf_cdna_manual.fa", "r"), "fasta")
        test_cds = SeqIO.parse(open("./data/Dlg1-Braf_cds_manual.fa", "r"), "fasta")

        expected_transcript_combinations = [
            "ENSMUST00000100001_ENSMUST00000002487",
            "ENSMUST00000064477_ENSMUST00000002487",
            "ENSMUST00000115205_ENSMUST00000002487",
            "ENSMUST00000023454_ENSMUST00000002487",
            "ENSMUST00000115201_ENSMUST00000002487",
            "ENSMUST00000132176_ENSMUST00000002487",
        ]

        assert (
            len(set(fusion.transcripts.keys()).intersection(set(expected_transcript_combinations)))
            == 6
        ), "Test 1: unexpected number protein coding transcripts."

        for seq in test_cdna:
            trans = fusion.transcripts[str(seq.id)]
            assert seq.seq == trans.cdna.seq, f"cDNA is wrongly predicted: {seq.id}"

        for seq in test_cds:
            trans = fusion.transcripts[str(seq.id)]
            assert seq.seq == trans.cds.seq, f"cds is wrongly predicted: {seq.id}"

    def test_2(self):
        """
        Test CDS correct for junction within the exon (not on boundary) for two
        genes on reverse strand
        """

        fusion = model.Fusion(
            gene5prime="ENSMUSG00000002413",
            gene5primejunction=39725110,
            gene3prime="ENSMUSG00000002413",
            gene3primejunction=39610402,
            db=db,
            pyensembl_data=data,
            protein_databases=["pfam", "tmhmm"],
            noncanonical=True,
        )

        cds = (
            "ATGGCGGCGCTGAGTGGCGGCGGTGGCAGCAGCAGCGGTGGCGGCGGCGGCGGTGGCGGCGGCGG"
            + "TGGCGGTGGCGACGGCGGCGGCGGCGCCGAGCAGGGCCAGGCTCTGTTCAATGGCGACATGGAGC"
            + "CGGAGGCCGGCGCTGGCGCCGCGGCCTCTTCGGCTGCGGACCCGGCCATTCCTGAAGAATTTGCAGCCTTCAAGTAG"
        )

        assert (
            str(fusion.transcripts["ENSMUST00000002487_ENSMUST00000002487"].cds.seq) == cds
        ), "Test 2: CDS wrong"

    def test_3(self):
        """
        Test CDS correct for junction within the exon (not on boundary) for one gene
        one the forward and one gene on the reverse strand
        """

        fusion = model.Fusion(
            gene5prime="ENSMUSG00000022770",
            gene5primejunction=31664869,
            gene3prime="ENSMUSG00000002413",
            gene3primejunction=39610402,
            db=db,
            pyensembl_data=data,
            protein_databases=["pfam", "tmhmm"],
            noncanonical=True,
        )

        cds = "ATGCCGGTCCGGAAGCAAGAATTTGCAGCCTTCAAGTAG"

        assert (
            str(fusion.transcripts["ENSMUST00000064477_ENSMUST00000002487"].cds.seq) == cds
        ), "Test 3: CDS wrong"

    def test_mouse_4(self):
        """
        Test cDNA correctly produced for junctions being in UTRs
        """

        fusion = model.Fusion(
            gene5prime="ENSMUSG00000022770",
            gene5primejunction=31664851,
            gene3prime="ENSMUSG00000022770",
            gene3primejunction=31873343,
            db=db,
            pyensembl_data=data,
            protein_databases=["pfam", "tmhmm"],
            noncanonical=True,
        )

        cdna = (
            "GGGGGTGCGGCCGCCGAAGGGGGAGCTCCTCCCCCGTCCCCTCACCCCCTCAGCTGAGCT"
            + "CGGGGCGGGGCGGGGTACGTGGAGCGGGGCCGGGCGGGGAAGCTGCTCCGAGTCCGGCCG"
            + "GAGCGCACCCGGGGCGCCCGCGTACGCCGCTCGCGGGAACTTTGCGGCGGAGCCGCAGGT"
            + "GTGGAGGCCGCGGAGGGGGGTGCATGAGCGGCGCGGAGAGCGGCGGCTGTCCGGTCCGGC"
            + "CCCTGCTGGAGTCGCCGCCGGGAGGAGACGAACGAGGAACCAG"
            + "GTGTGTGCCGCCTTCCTGATTCTGGAGAAAA"
            + "AAAA"
        )

        assert (
            str(fusion.transcripts["ENSMUST00000064477_ENSMUST00000064477"].cdna.seq) == cdna
        ), "Test 4: cDNA wrong"

    def test_mouse_5(self):
        """Test transcript constructed correctly (mouse)"""

        fusion = model.Fusion(
            gene5prime="ENSMUSG00000022770",
            gene5primejunction=31664850,
            gene3prime="ENSMUSG00000002413",
            gene3primejunction=39603240,
            db=db,
            pyensembl_data=data,
            protein_databases=["pfam", "tmhmm"],
            noncanonical=True,
        )

        cdna = (
            "GGGGGTGCGGCCGCCGAAGGGGGAGCTCCTCCCCCGTCCCCTCACCCCCTCAGCTGAGCT"
            + "CGGGGCGGGGCGGGGTACGTGGAGCGGGGCCGGGCGGGGAAGCTGCTCCGAGTCCGGCCG"
            + "GAGCGCACCCGGGGCGCCCGCGTACGCCGCTCGCGGGAACTTTGCGGCGGAGCCGCAGGT"
            + "GTGGAGGCCGCGGAGGGGGGTGCATGAGCGGCGCGGAGAGCGGCGGCTGTCCGGTCCGGC"
            + "CCCTGCTGGAGTCGCCGCCGGGAGGAGACGAACGAGGAACCAG"
            + "GTGTGTGCCGCCTTCCTGATTCTGGAGAAA"
            + "GAAA"
        )

        assert (
            str(fusion.transcripts["ENSMUST00000064477_ENSMUST00000002487"].cdna.seq) == cdna
        ), "Test 10: incorrect cDNA"


class TestEffectPrediciton(unittest.TestCase):
    """Test correctly predict junction location annotation and fusion effect."""

    def test_mouse_1(self):
        """
        Test that AGFusion determines if the effect on each individual transcript
        """

        fusion = model.Fusion(
            gene5prime="ENSMUSG00000022770",
            gene5primejunction=31664852,
            gene3prime="ENSMUSG00000002413",
            gene3primejunction=39651764,
            db=db,
            pyensembl_data=data,
            protein_databases=["pfam", "tmhmm"],
            noncanonical=True,
        )

        assert (
            fusion.transcripts["ENSMUST00000064477_ENSMUST00000002487"].effect_5prime
            == "CDS (start)"
        ), "Test 5: not CDS start"
        assert (
            fusion.transcripts["ENSMUST00000064477_ENSMUST00000002487"].effect_3prime == "CDS"
        ), "Test 5: not CDS"

    def test_mouse_2(self):
        """Test junction location annotated"""

        fusion = model.Fusion(
            gene5prime="ENSMUSG00000022770",
            gene5primejunction=31664851,
            gene3prime="ENSMUSG00000002413",
            gene3primejunction=39610381,
            db=db,
            pyensembl_data=data,
            protein_databases=["pfam", "tmhmm"],
            noncanonical=True,
        )

        assert (
            fusion.transcripts["ENSMUST00000064477_ENSMUST00000002487"].effect_5prime
            == "5UTR (end)"
        ), "Test 6: Not found in 5'UTR-end"
        assert (
            fusion.transcripts["ENSMUST00000064477_ENSMUST00000002487"].effect_3prime
            == "3UTR (start)"
        ), "Test 6: Not found in at 3'UTR beginning"

    def test_mouse_3(self):
        """Test junction location annotated"""

        fusion = model.Fusion(
            gene5prime="ENSMUSG00000022770",
            gene5primejunction=31664850,
            gene3prime="ENSMUSG00000002413",
            gene3primejunction=39610381,
            db=db,
            pyensembl_data=data,
            protein_databases=["pfam", "tmhmm"],
            noncanonical=True,
        )
        assert (
            fusion.transcripts["ENSMUST00000064477_ENSMUST00000002487"].effect_5prime == "5UTR"
        ), "Test 7: Not found in 5'UTR"
        assert (
            fusion.transcripts["ENSMUST00000064477_ENSMUST00000002487"].effect_3prime
            == "3UTR (start)"
        ), "Test 7: Not found in at 3'UTR beginning"

    def test_mouse_4(self):
        """Test junction location annotated"""

        fusion = model.Fusion(
            gene5prime="ENSMUSG00000022770",
            gene5primejunction=31664851,
            gene3prime="ENSMUSG00000022770",
            gene3primejunction=31871782,
            db=db,
            pyensembl_data=data,
            protein_databases=["pfam", "tmhmm"],
            noncanonical=True,
        )

        assert (
            fusion.transcripts["ENSMUST00000064477_ENSMUST00000064477"].effect_5prime
            == "5UTR (end)"
        ), "Test 8: Not found in 5'UTR-end"
        assert (
            fusion.transcripts["ENSMUST00000064477_ENSMUST00000064477"].effect_3prime
            == "3UTR (start)"
        ), "Test 8: Not found in at 3'UTR beginning"

    def test_mouse_5(self):
        """Test junction location annotated"""

        fusion = model.Fusion(
            gene5prime="ENSMUSG00000002413",
            gene5primejunction=39725296,
            gene3prime="ENSMUSG00000002413",
            gene3primejunction=39610381,
            db=db,
            pyensembl_data=data,
            protein_databases=["pfam", "tmhmm"],
            noncanonical=True,
        )

        assert (
            fusion.transcripts["ENSMUST00000002487_ENSMUST00000002487"].effect_5prime
            == "5UTR (end)"
        ), "Test 9: Not found in 5'UTR-end"
        assert (
            fusion.transcripts["ENSMUST00000002487_ENSMUST00000002487"].effect_3prime
            == "3UTR (start)"
        ), "Test 9: Not found in at 3'UTR beginning"

    def test_mouse_6(self):
        """Test junction location annotated"""

        fusion = model.Fusion(
            gene5prime="ENSMUSG00000022770",
            gene5primejunction=31743271,
            gene3prime="ENSMUSG00000002413",
            gene3primejunction=39665003,
            db=db,
            pyensembl_data=data,
            protein_databases=["pfam", "tmhmm"],
            noncanonical=True,
        )
        transcript = fusion.transcripts["ENSMUST00000023454_ENSMUST00000002487"]

        assert (
            transcript.effect_5prime == "intron (cds)"
        ), f"Test 11: incorrect 5' effect: {transcript.effect_5prime}"
        assert (
            transcript.effect_3prime == "intron (cds)"
        ), f"Test 11: incorrect 3' effect: {transcript.effect_3prime}"

    def test_mouse_7(self):
        """Test junction location annotated"""

        fusion = model.Fusion(
            gene5prime="ENSMUSG00000022770",
            gene5primejunction=31664820,
            gene3prime="ENSMUSG00000002413",
            gene3primejunction=39610405,
            db=db,
            pyensembl_data=data,
            protein_databases=["pfam", "tmhmm"],
            noncanonical=True,
        )
        transcript = fusion.transcripts["ENSMUST00000023454_ENSMUST00000002487"]

        assert (
            transcript.effect_5prime == "intron (before cds)"
        ), f"Test 12: incorrect 5' effect: {transcript.effect_5prime}"
        assert (
            transcript.effect_3prime == "intron (cds)"
        ), f"Test 12: incorrect 3' effect: {transcript.effect_3prime}"


class TestFusionCatcher(unittest.TestCase):
    """Test parse FusionCatcher parse."""

    def test_1(self):
        """Test basic parsing."""

        all_fusions = [
            "Adamts9_Ano2",
            "Trp53_Sat2",
            "1700112E06Rik_Runx1",
            "Runx1_1700112E06Rik",
            "Rell1_Lhfpl3",
            "Phc1_Smarca2",
            "Lrrc8d_Gbp11",
            "C920009B18Rik_H60b",
        ]
        for fusion in parsers.parsers["fusioncatcher"](
            "./data/FusionsFindingAlgorithms/FusionCatcher/final-list_candidate-fusion-genes.txt",
            db.logger,
        ):
            fusion = model.Fusion(
                gene5prime=fusion["gene5prime"],
                gene5primejunction=fusion["gene5prime_junction"],
                gene3prime=fusion["gene3prime"],
                gene3primejunction=fusion["gene3prime_junction"],
                db=db,
                pyensembl_data=data,
                protein_databases=["pfam"],
                noncanonical=False,
            )
            assert fusion.name in all_fusions, f"{fusion.name} not in list!"


class TestSTARFusion(unittest.TestCase):
    """Test parse STAR Fusion"""

    def test_1(self):
        """Test basic parsing."""

        all_fusions = [
            "Adamts9_Ano2",
            "Trp53_Sat2",
            "1700112E06Rik_Runx1",
            "Runx1_1700112E06Rik",
            "Rell1_Lhfpl3",
            "Phc1_Smarca2",
            "Lrrc8d_Gbp11",
            "C920009B18Rik_H60b",
        ]
        for fusion in parsers.parsers["fusioncatcher"](
            "./data/FusionsFindingAlgorithms/FusionCatcher/final-list_candidate-fusion-genes.txt",
            db.logger,
        ):
            fusion = model.Fusion(
                gene5prime=fusion["gene5prime"],
                gene5primejunction=fusion["gene5prime_junction"],
                gene3prime=fusion["gene3prime"],
                gene3primejunction=fusion["gene3prime_junction"],
                db=db,
                pyensembl_data=data,
                protein_databases=["pfam"],
                noncanonical=False,
            )
            assert fusion.name in all_fusions, f"{fusion.name} not in list!"


if __name__ == "__main__":
    unittest.main()



"""
AGFusion is a python package to annotate and visualize gene fusions.
"""

import re

from setuptools import find_packages, setup

VERSIONFILE = "agfusion/version.py"
version = open(VERSIONFILE, "rt", encoding="utf8").read()
VSRE = r"^__version__ = ['\"]([^'\"]*)['\"]"
mo = re.search(VSRE, version, re.M)
if mo:
    version = mo.group(1)
else:
    raise RuntimeError(f"Unable to find version string in {VERSIONFILE}.")

README = open("README.md", encoding="utf8").read()

setup(
    # version=version,
    version="1.4.1",
    name="agfusion",
    packages=find_packages(),
    description="Python package to annotate and visualize gene fusions.",
    author="Charles Murphy",
    author_email="murphy.charlesj@gmail.com",
    license="MIT",
    url="https://github.com/murphycj/AGFusion",
    long_description=README,
    long_description_content_type="text/markdown",
    include_package_data=True,
    scripts=["bin/agfusion"],
    install_requires=[
        "matplotlib>=1.5.0",
        "pandas>=0.18.1",
        "biopython>=1.67",
        "future>=0.16.0",
        "pyensembl>=1.1.0",
    ],
)
